import os

from ms1_id.msi.calculate_mz_cor_parallel import calc_all_mz_correlations
from ms1_id.msi.export_msi import write_ms1_id_results
from ms1_id.msi.group_mz_cor_parallel import generate_pseudo_ms2
from ms1_id.msi.process_msi_data import process_ms_imaging_data
from ms1_id.msi.reverse_matching_parallel import validate_library_path, ms1_id_annotation


def ms1id_imaging_workflow(file_path, library_path, n_processes=None,
                           mass_detect_int_tol=None, noise_detection='moving_average',
                           sn_factor=5.0, centroided=False,
                           mz_bin_size=0.01,
                           min_overlap=10, min_correlation=0.85, max_cor_depth=1,
                           library_search_mztol=0.05,
                           ms1id_score_cutoff=0.7, ms1id_min_matched_peak=4,
                           ms1id_min_spec_usage=0.10, max_prec_rel_int_in_other_ms2=0.05):
    file_dir = os.path.dirname(file_path)
    file_name = os.path.splitext(os.path.basename(file_path))[0]

    # validate library_path
    library_path = validate_library_path(library_path)

    # make a result folder of file_name
    result_folder = os.path.join(file_dir, file_name)
    os.makedirs(result_folder, exist_ok=True)

    print(f"Processing {file_name}")
    mz_values, intensity_matrix, coordinates, ion_mode = process_ms_imaging_data(
        file_path,
        os.path.splitext(file_path)[0] + '.ibd',
        mz_bin_size=mz_bin_size,
        mass_detect_int_tol=mass_detect_int_tol,
        noise_detection=noise_detection,
        sn_factor=sn_factor,
        centroided=centroided,
        n_processes=n_processes,
        save=True, save_dir=result_folder
    )

    print(f"Calculating ion image correlations for {file_name}")
    cor_matrix = calc_all_mz_correlations(intensity_matrix,
                                          min_overlap=min_overlap,
                                          min_cor=min_correlation,
                                          n_processes=n_processes,
                                          save=True,
                                          save_dir=result_folder)

    print(f"Generating pseudo MS2 spectra for {file_name}")
    pseudo_ms2 = generate_pseudo_ms2(mz_values, intensity_matrix, cor_matrix,
                                     n_processes=n_processes,
                                     min_cluster_size=ms1id_min_matched_peak + 1,
                                     min_cor=min_correlation,
                                     max_cor_depth=max_cor_depth,
                                     save=True,
                                     save_dir=result_folder)

    print(f"Annotating pseudo MS2 spectra for {file_name}")
    pseudo_ms2 = ms1_id_annotation(pseudo_ms2, library_path, n_processes=None,
                                   mz_tol=library_search_mztol,
                                   ion_mode=ion_mode,
                                   score_cutoff=ms1id_score_cutoff,
                                   min_matched_peak=ms1id_min_matched_peak,
                                   min_spec_usage=ms1id_min_spec_usage,
                                   max_prec_rel_int_in_other_ms2=max_prec_rel_int_in_other_ms2,
                                   save=True,
                                   save_dir=result_folder)

    print(f"Writing results for {file_name}")
    write_ms1_id_results(pseudo_ms2, save=True, save_dir=result_folder)

    return



import pickle
from collections import Counter

import pandas as pd
import requests
from rdkit import Chem
from rdkit import DataStructs
from rdkit.Avalon import pyAvalonTools
from rdkit.Chem import AllChem
from tqdm import tqdm


def inchikey_to_smiles(inchikey):
    """
    Convert an InChIKey to SMILES using the PubChem PUG REST API.
    If the original InChIKey isn't found, it tries with a modified version. -UHFFFAOYSA-N
    """
    base_url = "https://pubchem.ncbi.nlm.nih.gov/rest/pug"

    def fetch_cid(key):
        cid_url = f"{base_url}/compound/inchikey/{key}/cids/JSON"
        response = requests.get(cid_url)
        if response.status_code != 200:
            return None
        data = response.json()
        return data.get("IdentifierList", {}).get("CID", [None])[0]

    def fetch_smiles(cid):
        smiles_url = f"{base_url}/compound/cid/{cid}/property/IsomericSMILES/JSON"
        response = requests.get(smiles_url)
        if response.status_code != 200:
            return None
        data = response.json()
        return data.get("PropertyTable", {}).get("Properties", [{}])[0].get("IsomericSMILES")

    # Try with original InChIKey
    cid = fetch_cid(inchikey)

    if cid is None:
        # If not found, modify InChIKey and try again
        modified_inchikey = inchikey[:14] + "-UHFFFAOYSA-N"
        print(f"Original InChIKey not found. Trying modified InChIKey: {modified_inchikey}")
        cid = fetch_cid(modified_inchikey)

    if cid is None:
        print(f"No CID found for InChIKey {inchikey} or its modified version")
        return None

    smiles = fetch_smiles(cid)
    if smiles is None:
        print(f"No SMILES found for CID {cid}")
        return None

    return smiles


def calculate_similarity_metrics(smiles1, smiles2):
    # Convert SMILES to RDKit molecules
    try:
        mol1 = Chem.MolFromSmiles(smiles1)
        mol2 = Chem.MolFromSmiles(smiles2)
    except:
        return None

    if mol1 is None or mol2 is None:
        return None

    # Remove stereochemistry
    Chem.RemoveStereochemistry(mol1)
    Chem.RemoveStereochemistry(mol2)

    # Generate fingerprints for all types
    morgan_fp1 = AllChem.GetMorganFingerprintAsBitVect(mol1, 2, nBits=2048)
    morgan_fp2 = AllChem.GetMorganFingerprintAsBitVect(mol2, 2, nBits=2048)

    rdkit_fp1 = Chem.RDKFingerprint(mol1)
    rdkit_fp2 = Chem.RDKFingerprint(mol2)

    maccs_fp1 = AllChem.GetMACCSKeysFingerprint(mol1)
    maccs_fp2 = AllChem.GetMACCSKeysFingerprint(mol2)

    # Add PubChem fingerprint
    pubchem_fp1 = AllChem.GetHashedAtomPairFingerprintAsBitVect(mol1, nBits=881)
    pubchem_fp2 = AllChem.GetHashedAtomPairFingerprintAsBitVect(mol2, nBits=881)

    # Add Avalon fingerprint
    avalon_fp1 = pyAvalonTools.GetAvalonFP(mol1)
    avalon_fp2 = pyAvalonTools.GetAvalonFP(mol2)

    # Calculate similarities for each fingerprint type
    morgan_sim = DataStructs.TanimotoSimilarity(morgan_fp1, morgan_fp2)
    rdkit_sim = DataStructs.TanimotoSimilarity(rdkit_fp1, rdkit_fp2)
    maccs_sim = DataStructs.TanimotoSimilarity(maccs_fp1, maccs_fp2)
    pubchem_sim = DataStructs.TanimotoSimilarity(pubchem_fp1, pubchem_fp2)
    avalon_sim = DataStructs.TanimotoSimilarity(avalon_fp1, avalon_fp2)

    return {
        'morgan': morgan_sim,
        'rdkit': rdkit_sim,
        'maccs': maccs_sim,
        'pubchem': pubchem_sim,
        'avalon': avalon_sim
    }


def gen_unique_smiles_for_df():
    df = pd.read_pickle('data/aligned_feature_table_all.pkl')

    all_inchikeys = set()

    # cols: InChIKey, ms1_dda_inchikeys, ms1_0ev_inchikeys, ms1_10ev_inchikeys, ms1_20ev_inchikeys
    for col in ['InChIKey', 'ms1_dda_inchikeys', 'ms1_0ev_inchikeys', 'ms1_10ev_inchikeys', 'ms1_20ev_inchikeys']:
        print(f"Processing column: {col}")
        # Drop NA values and get unique values
        values = df[col].dropna()

        for value in values:
            if isinstance(value, list):
                all_inchikeys.update(value)
            else:
                all_inchikeys.add(value)

    # remove nan inchikey and stereochemistry
    all_inchikeys = [i[:14] for i in all_inchikeys if not pd.isnull(i)]
    all_inchikeys = list(set(all_inchikeys))

    print(f"Total unique InChIKeys: {len(all_inchikeys)}")

    ########################################################
    # Convert InChIKeys to SMILES, using GNPS DB
    # load gnps db metadata df
    gnps = pd.read_pickle('/data/ms2db/gnps.pkl')
    gnps = gnps[['inchikey_14', 'smiles']]
    # remove nan rows
    gnps = gnps[(gnps['inchikey_14'].notnull()) & (gnps['smiles'].notnull()) & (gnps['smiles'] != 'N/A')].reset_index(
        drop=True)
    # Group by inchikey_14 and find the most common SMILES for each
    gnps_grouped = gnps.groupby('inchikey_14')['smiles'].agg(lambda x: Counter(x).most_common(1)[0][0]).reset_index()

    # Convert to dictionary for faster lookup
    gnps_dict = dict(zip(gnps_grouped['inchikey_14'], gnps_grouped['smiles']))

    # create a dictionary to store InChIKey to SMILES
    inchikey_to_smiles_dict = {}
    for inchikey in tqdm(all_inchikeys):
        if inchikey in gnps_dict:
            inchikey_to_smiles_dict[inchikey] = gnps_dict[inchikey]
        else:
            print(f"Could not find SMILES for InChIKey: {inchikey}")
            smiles = inchikey_to_smiles(str(inchikey) + '-UHFFFAOYSA-N')
            if smiles is not None:
                inchikey_to_smiles_dict[inchikey] = smiles
            else:
                inchikey_to_smiles_dict[inchikey] = None

    # save, inchikeys are first 14 characters
    pickle.dump(inchikey_to_smiles_dict, open('data/inchikey_to_smiles_dict.pkl', 'wb'))
    print(f"Total unique SMILES: {len(inchikey_to_smiles_dict)}")


def get_structure_similarity(ms2_inchikey, inchikey_ls, inchikey_to_smiles_dict):
    if not inchikey_ls:
        return None

    if ms2_inchikey in inchikey_ls:  # identical compounds found
        return [1.0, 1.0, 1.0, 1.0, 1.0]

    ms2_smiles = inchikey_to_smiles_dict[ms2_inchikey]
    if ms2_smiles is None:
        return None

    smiles_ls = [inchikey_to_smiles_dict[inchikey] for inchikey in inchikey_ls]
    smiles_ls = [smiles for smiles in smiles_ls if smiles is not None]

    if len(smiles_ls) == 0:
        return None

    for smiles in smiles_ls:
        similarity = calculate_similarity_metrics(ms2_smiles, smiles)
        if similarity is not None:
            return [similarity['morgan'], similarity['rdkit'], similarity['maccs'], similarity['pubchem'], similarity['avalon']]

    return None


def main():
    df = pd.read_pickle('data/aligned_feature_table_all.pkl')
    inchikey_to_smiles_dict = pickle.load(open('data/inchikey_to_smiles_dict.pkl', 'rb'))

    df['ms1_dda_struct_sim'] = [None] * df.shape[0]
    df['ms1_0ev_struct_sim'] = [None] * df.shape[0]
    df['ms1_10ev_struct_sim'] = [None] * df.shape[0]
    df['ms1_20ev_struct_sim'] = [None] * df.shape[0]

    for idx, row in tqdm(df.iterrows(), total=df.shape[0]):
        if pd.isnull(row['InChIKey']):
            continue

        ms2_inchikey = row['InChIKey'][:14]
        ms1_dda_inchikeys = [i[:14] for i in row['ms1_dda_inchikeys'] if pd.notnull(i)] if row['ms1_dda_inchikeys'] else []
        ms1_0ev_inchikeys = [i[:14] for i in row['ms1_0ev_inchikeys'] if pd.notnull(i)] if row['ms1_0ev_inchikeys'] else []
        ms1_10ev_inchikeys = [i[:14] for i in row['ms1_10ev_inchikeys'] if pd.notnull(i)] if row['ms1_10ev_inchikeys'] else []
        ms1_20ev_inchikeys = [i[:14] for i in row['ms1_20ev_inchikeys'] if pd.notnull(i)] if row['ms1_20ev_inchikeys'] else []

        ms1_dda_inchikeys = list(set(ms1_dda_inchikeys))
        ms1_0ev_inchikeys = list(set(ms1_0ev_inchikeys))
        ms1_10ev_inchikeys = list(set(ms1_10ev_inchikeys))
        ms1_20ev_inchikeys = list(set(ms1_20ev_inchikeys))

        df.at[idx, 'ms1_dda_struct_sim'] = get_structure_similarity(ms2_inchikey, ms1_dda_inchikeys,
                                                                    inchikey_to_smiles_dict)
        df.at[idx, 'ms1_0ev_struct_sim'] = get_structure_similarity(ms2_inchikey, ms1_0ev_inchikeys,
                                                                    inchikey_to_smiles_dict)
        df.at[idx, 'ms1_10ev_struct_sim'] = get_structure_similarity(ms2_inchikey, ms1_10ev_inchikeys,
                                                                     inchikey_to_smiles_dict)
        df.at[idx, 'ms1_20ev_struct_sim'] = get_structure_similarity(ms2_inchikey, ms1_20ev_inchikeys,
                                                                     inchikey_to_smiles_dict)

    df.to_pickle('data/aligned_feature_table_all_struct_sim.pkl')


def print_stats():
    df = pd.read_pickle('data/aligned_feature_table_all_struct_sim.pkl')

    print('ms1_dda:')
    print('morgan:', df['ms1_dda_struct_sim'].apply(lambda x: x[0] if x is not None else None).mean())
    print('rdkit:', df['ms1_dda_struct_sim'].apply(lambda x: x[1] if x is not None else None).mean())
    print('maccs:', df['ms1_dda_struct_sim'].apply(lambda x: x[2] if x is not None else None).mean())
    print('pubchem:', df['ms1_dda_struct_sim'].apply(lambda x: x[3] if x is not None else None).mean())
    print('avalon:', df['ms1_dda_struct_sim'].apply(lambda x: x[4] if x is not None else None).mean())

    print('morgan:', df['ms1_dda_struct_sim'].apply(lambda x: x[0] if x is not None else None).median())
    print('rdkit:', df['ms1_dda_struct_sim'].apply(lambda x: x[1] if x is not None else None).median())
    print('maccs:', df['ms1_dda_struct_sim'].apply(lambda x: x[2] if x is not None else None).median())
    print('pubchem:', df['ms1_dda_struct_sim'].apply(lambda x: x[3] if x is not None else None).median())
    print('avalon:', df['ms1_dda_struct_sim'].apply(lambda x: x[4] if x is not None else None).median())

    print('ms1_0ev:')
    print('morgan:', df['ms1_0ev_struct_sim'].apply(lambda x: x[0] if x is not None else None).mean())
    print('rdkit:', df['ms1_0ev_struct_sim'].apply(lambda x: x[1] if x is not None else None).mean())
    print('maccs:', df['ms1_0ev_struct_sim'].apply(lambda x: x[2] if x is not None else None).mean())
    print('pubchem:', df['ms1_0ev_struct_sim'].apply(lambda x: x[3] if x is not None else None).mean())
    print('avalon:', df['ms1_0ev_struct_sim'].apply(lambda x: x[4] if x is not None else None).mean())

    print('morgan:', df['ms1_0ev_struct_sim'].apply(lambda x: x[0] if x is not None else None).median())
    print('rdkit:', df['ms1_0ev_struct_sim'].apply(lambda x: x[1] if x is not None else None).median())
    print('maccs:', df['ms1_0ev_struct_sim'].apply(lambda x: x[2] if x is not None else None).median())
    print('pubchem:', df['ms1_0ev_struct_sim'].apply(lambda x: x[3] if x is not None else None).median())
    print('avalon:', df['ms1_0ev_struct_sim'].apply(lambda x: x[4] if x is not None else None).median())

    print('ms1_10ev:')
    print('morgan:', df['ms1_10ev_struct_sim'].apply(lambda x: x[0] if x is not None else None).mean())
    print('rdkit:', df['ms1_10ev_struct_sim'].apply(lambda x: x[1] if x is not None else None).mean())
    print('maccs:', df['ms1_10ev_struct_sim'].apply(lambda x: x[2] if x is not None else None).mean())
    print('pubchem:', df['ms1_10ev_struct_sim'].apply(lambda x: x[3] if x is not None else None).mean())
    print('avalon:', df['ms1_10ev_struct_sim'].apply(lambda x: x[4] if x is not None else None).mean())

    print('morgan:', df['ms1_10ev_struct_sim'].apply(lambda x: x[0] if x is not None else None).median())
    print('rdkit:', df['ms1_10ev_struct_sim'].apply(lambda x: x[1] if x is not None else None).median())
    print('maccs:', df['ms1_10ev_struct_sim'].apply(lambda x: x[2] if x is not None else None).median())
    print('pubchem:', df['ms1_10ev_struct_sim'].apply(lambda x: x[3] if x is not None else None).median())
    print('avalon:', df['ms1_10ev_struct_sim'].apply(lambda x: x[4] if x is not None else None).median())

    print('ms1_20ev:')
    print('morgan:', df['ms1_20ev_struct_sim'].apply(lambda x: x[0] if x is not None else None).mean())
    print('rdkit:', df['ms1_20ev_struct_sim'].apply(lambda x: x[1] if x is not None else None).mean())
    print('maccs:', df['ms1_20ev_struct_sim'].apply(lambda x: x[2] if x is not None else None).mean())
    print('pubchem:', df['ms1_20ev_struct_sim'].apply(lambda x: x[3] if x is not None else None).mean())
    print('avalon:', df['ms1_20ev_struct_sim'].apply(lambda x: x[4] if x is not None else None).mean())

    print('morgan:', df['ms1_20ev_struct_sim'].apply(lambda x: x[0] if x is not None else None).median())
    print('rdkit:', df['ms1_20ev_struct_sim'].apply(lambda x: x[1] if x is not None else None).median())
    print('maccs:', df['ms1_20ev_struct_sim'].apply(lambda x: x[2] if x is not None else None).median())
    print('pubchem:', df['ms1_20ev_struct_sim'].apply(lambda x: x[3] if x is not None else None).median())
    print('avalon:', df['ms1_20ev_struct_sim'].apply(lambda x: x[4] if x is not None else None).median())


import matplotlib.pyplot as plt
import seaborn as sns


def plot_density():
    # Load the data
    df = pd.read_pickle('data/aligned_feature_table_all_struct_sim.pkl')

    # Define colors
    colors = ['#56648a', '#6280a5', '#8ca5c0', '#8d7e95', '#ca9a96', '#facaa9']

    # Set the font to Arial
    plt.rcParams['font.family'] = 'Arial'

    # Create the plot
    fig, ax = plt.subplots(figsize=(5.2, 2.0))

    # Prepare legend labels with counts
    legend_labels = []
    medians = []
    for i, col in enumerate(['ms1_dda_struct_sim', 'ms1_0ev_struct_sim', 'ms1_10ev_struct_sim', 'ms1_20ev_struct_sim']):
        data = df[col].apply(lambda x: x[2] if x is not None else None).dropna()
        count = len(data)
        label = f'MS1 ({col.split("_")[1].upper().replace("EV", " eV")}) (n = {count})'
        # label = f'MS1 ({col.split("_")[1].upper().replace("EV", " eV")})'
        legend_labels.append(label)

        sns.kdeplot(data, fill=False, color=colors[i], ax=ax, bw_adjust=0.2)

        median = data.median()
        medians.append(median)

    # Set labels and title
    ax.set_xlabel('Structure similarity (compared to MS/MS annotations)', color='0.2', fontsize=13, labelpad=3)
    ax.set_ylabel('Density', color='0.2', fontsize=13, labelpad=4)

    # Adjust tick parameters
    ax.tick_params(axis='x', which='major', labelsize=10, colors='0.2', length=2, pad=3)
    ax.tick_params(axis='y', which='major', labelsize=10, colors='0.2', length=2, pad=3)

    # Set frame color
    for spine in ax.spines.values():
        spine.set_edgecolor('0.4')

    # Create legend inside the plot
    legend = ax.legend(legend_labels, loc='upper left', bbox_to_anchor=(0.05, 0.95), borderaxespad=0, fontsize=10)
    legend.get_frame().set_linewidth(0.15)  # Set a thin line around the legend
    legend.get_frame().set_facecolor('none')  # Make legend background transparent

    # Change legend text color to 0.2
    for text in legend.get_texts():
        text.set_color('0.2')

    # Add median value texts
    y_positions = [8, 6.5, 5, 3.5]  # Adjust these values to position the text vertically
    for i, (median, color) in enumerate(zip(medians, colors)):
        ax.text(0.85, y_positions[i], f'Median: {median:.2f}', color=color,
                ha='center', va='center', fontsize=9.5)

    # Set x-axis limits
    ax.set_xlim(0, 1.002)

    # Adjust layout to prevent cutting off labels
    plt.tight_layout()

    # save
    plt.savefig('data/structure_similarity_density.svg', transparent=True)

    # Show the plot
    plt.show()


def plot_similarity_pie_charts():
    # Load the data
    df = pd.read_pickle('data/aligned_feature_table_all_struct_sim.pkl')

    # Set the font to Arial
    plt.rcParams['font.family'] = 'Arial'

    # Create the plot
    fig, axs = plt.subplots(1, 4, figsize=(5, 2.0))

    # Define the ranges
    # colors = ['#56648a', '#6280a5', '#8ca5c0', '#8d7e95']
    # ranges = [(0, 0.5), (0.5, 0.7), (0.7, 0.95), (0.95, 1.0)]
    # range_labels = ['0.00-0.50', '0.50-0.70', '0.70-0.95', '0.95-1.00']

    colors = ['#56648a', '#6280a5', '#8d7e95']
    ranges = [(0, 0.5), (0.5, 0.7), (0.7, 1.0)]
    range_labels = ['0.00-0.50', '0.50-0.70', '0.70-1.00']

    for i, col in enumerate(['ms1_dda_struct_sim', 'ms1_0ev_struct_sim', 'ms1_10ev_struct_sim', 'ms1_20ev_struct_sim']):
        data = df[col].apply(lambda x: x[2] if x is not None else None).dropna()

        # Calculate percentages for each range
        percentages = [
            (data[(data > r[0]) & (data <= r[1])].count() / len(data)) * 100
            for r in ranges
        ]

        # Create pie chart without labels
        axs[i].pie(percentages, colors=colors, autopct=lambda pct: f'{pct:.1f}%' if pct > 5 else '',
                   startangle=90, textprops={'fontsize': 6.5, 'color': '0.95'})
        axs[i].set_title(f'MS1 ({col.split("_")[1].upper().replace("EV", " eV")})\nn = {len(data)}', fontsize=10)

    # Create a common legend
    fig.legend(range_labels, loc='lower center', bbox_to_anchor=(0.5, 0.1), frameon=False,
               ncol=4, fancybox=False, shadow=False, fontsize=9)

    # Adjust layout
    plt.tight_layout()

    plt.savefig('data/structure_similarity_pie_charts.svg', transparent=True)

    # Show the plot
    plt.show()


if __name__ == "__main__":
    # gen_unique_smiles_for_df()
    # #
    # main()

    print_stats()

    plot_density()
    #
    # plot_similarity_pie_charts()



import os
import argparse
from ms1_id.lcms.reverse_matching import prepare_ms2_lib
from ms1_id.lcms.main_lcms import ms1id_batch_mode as ms1id_lcms
from ms1_id.msi.main_msi import ms1id_single_file_batch as ms1id_msi
from ms1_id.annotate import annotate_pseudo_ms2_spec


def index_library(args):
    """Function to handle library indexing"""

    # Check if input file exists
    if not os.path.exists(args.ms2db):
        raise FileNotFoundError(f"MS2 database file not found: {args.ms2db}")

    print(f"Indexing library from: {args.ms2db}")
    # print(f"Using m/z tolerance: {args.mz_tol}")
    print(f"Peak scaling factor: {args.peak_scale_k}")
    print(f"Peak intensity power: {args.peak_intensity_power}")

    # Call your actual indexing function here
    prepare_ms2_lib(
        ms2db=args.ms2db,
        # mz_tol=args.mz_tol,
        peak_scale_k=args.peak_scale_k,
        peak_intensity_power=args.peak_intensity_power
    )


def annotate_spectra(args):
    """Function to handle spectra annotation"""

    # Check if input file exists
    if not os.path.exists(args.input_file):
        raise FileNotFoundError(f"Input file not found: {args.input_file}")

    # Verify library paths
    lib_ls = _verify_paths(args.libs)
    if not lib_ls:
        raise ValueError("No valid library paths provided")

    print(f"Annotating spectra from: {args.input_file}")
    print(f"Using libraries: {args.libs}")
    print(f"m/z tolerance: {args.mz_tol}")
    print(f"Ion mode: {args.ion_mode}")
    print(f"Score cutoff: {args.min_score}")
    print(f"Minimum matched peak: {args.min_matched_peak}")
    print(f"Minimum spectral usage: {args.min_spec_usage}")
    if args.output_folder is None:
        print("No output folder specified. Results will saved in the same folder as the input file.")
    else:
        print(f"Output folder: {args.output_folder}")

    annotate_pseudo_ms2_spec(
        ms2_file_path=args.input_file,
        library_ls=lib_ls,
        mz_tol=args.mz_tol,
        ion_mode=args.ion_mode,
        score_cutoff=args.min_score,
        min_matched_peak=args.min_matched_peak,
        min_spec_usage=args.min_spec_usage,
        save_dir=args.output_folder
    )


def run_lcms(args):

    # Check if project path exists
    if not os.path.exists(args.project_dir):
        raise FileNotFoundError(f"Project directory not found: {args.project_dir}")

    # Verify library paths
    ms1_libs = _verify_paths(args.ms1_id_libs)

    print(f"Running LC-MS data analysis in project directory: {args.project_dir}")
    print(f"Sample directory: {args.sample_dir}")

    if ms1_libs:
        print("MS1 ID enabled")
        print(f"Using MS1 ID libraries: {ms1_libs}")
    else:
        print("MS1 ID disabled, as no MS1 ID libraries are provided")

    if args.ms2_id_lib:
        print("MS2 ID enabled")
        print(f"Using MS2 ID libraries: {args.ms2_id_lib}")
    else:
        print("MS2 ID disabled, as no MS2 ID libraries are provided")

    print(f"Parallel processing: {args.parallel}")
    print(f"CPU ratio: {args.cpu_ratio}")

    # Run the analysis
    ms1id_lcms(
        project_path=args.project_dir,
        ms1id_library_path=ms1_libs,
        ms2id_library_path=args.ms2_id_lib,
        sample_dir=args.sample_dir,
        parallel=args.parallel,
        ms1_id=True if ms1_libs else False,
        ms2_id=True if args.ms2_id_lib else False,
        cpu_ratio=args.cpu_ratio,
        run_rt_correction=args.run_rt_correction,
        run_normalization=args.run_normalization,
        ms1_tol=args.ms1_tol,
        ms2_tol=args.ms2_tol,
        mass_detect_int_tol=args.mass_detect_int_tol,
        align_mz_tol=args.align_mz_tol,
        align_rt_tol=args.align_rt_tol,
        alignment_drop_by_fill_pct_ratio=args.alignment_drop_by_fill_pct_ratio,
        peak_cor_rt_tol=args.peak_cor_rt_tol,
        min_ppc=args.min_ppc,
        roi_min_length=args.roi_min_length,
        library_search_mztol=args.library_search_mztol,
        ms1id_score_cutoff=args.ms1id_score_cutoff,
        ms1id_min_matched_peak=args.ms1id_min_matched_peak,
        ms1id_min_spec_usage=args.ms1id_min_spec_usage,
        ms1id_max_prec_rel_int_in_other_ms2=args.ms1id_max_prec_rel_int_in_other_ms2,
        ms2id_score_cutoff=args.ms2id_score_cutoff,
        ms2id_min_matched_peak=args.ms2id_min_matched_peak
    )


def run_msi(args):

    # Check if project directory exists
    if not os.path.exists(args.project_dir):
        raise FileNotFoundError(f"Project directory not found: {args.project_dir}")

    # Verify library paths
    library_paths = _verify_paths(args.libs)

    print(f"Running MS imaging data analysis in project directory: {args.project_dir}")
    print(f"Using libraries: {library_paths}")
    print(f"Number of cores used: {args.n_cores}")

    # Run the analysis
    ms1id_msi(
        file_dir=args.project_dir,
        library_path=library_paths,
        n_processes=args.n_cores,
        mass_detect_int_tol=args.mass_detect_int_tol,
        noise_detection=args.noise_detection,
        sn_factor=args.sn_factor,
        centroided=args.centroided,
        mz_bin_size=args.mz_bin_size,
        min_overlap=args.min_overlap,
        min_correlation=args.min_correlation,
        max_cor_depth=args.max_cor_depth,
        library_search_mztol=args.library_search_mztol,
        ms1id_score_cutoff=args.ms1id_score_cutoff,
        ms1id_min_matched_peak=args.ms1id_min_matched_peak,
        ms1id_min_spec_usage=args.ms1id_min_spec_usage
    )


def _verify_paths(paths):
    """Verify that all paths exist"""
    if paths is None:
        return []
    for path in paths:
        if not os.path.exists(path):
            raise FileNotFoundError(f"Path not found: {path}")
    return paths


def main():
    parser = argparse.ArgumentParser(description='Annotate LC-MS1 data, MS imaging data or pseudo MS/MS spectra using reference MS/MS libraries')
    subparsers = parser.add_subparsers(dest='command', help='Commands')

    ####################
    # Subcommand for indexing library
    index_parser = subparsers.add_parser('index', help='Index MS/MS library from a msp or mgf file')
    index_parser.add_argument('--ms2db', type=str, required=True,
                        help='Path to MS/MS database file in mgf or msp format')
    # index_parser.add_argument('--mz_tol', type=float, default=0.05,
    #                     help='m/z tolerance for peak matching (default: 0.05)')
    index_parser.add_argument('--peak_scale_k', type=float, default=None,
                        help='Peak scaling factor. Set to None for no scaling (default: None)')
    index_parser.add_argument('--peak_intensity_power', type=float, default=1.0,
                        help='Peak intensity power. Use 0.5 for square root transformation (default: 1.0, no transformation)')

    ####################
    # Subcommand for annotation
    annotate_parser = subparsers.add_parser('annotate', help='Annotate pseudo MS/MS spectra in mgf format')
    annotate_parser.add_argument('--input_file', '-i', type=str, required=True,
                        help='Path to the MGF file containing pseudo MS/MS spectra')
    annotate_parser.add_argument('--libs', '-l', type=str, required=True, nargs='*', default=None,
                        help='One or more paths to indexed library files (.pkl). Paths with spaces should be quoted.')
    annotate_parser.add_argument('--output_folder', '-o', type=str, default=None,
                        help='Output folder for annotated results')
    annotate_parser.add_argument('--mz_tol', type=float, default=0.05,
                        help='m/z tolerance for peak matching (default: 0.05)')
    annotate_parser.add_argument('--ion_mode', type=str, default=None,
                        help='Ion mode, "positive" or "negative" (default: None)')
    annotate_parser.add_argument('--min_score', type=float, default=0.7,
                        help='Minimum score for matching (default: 0.7)')
    annotate_parser.add_argument('--min_matched_peak', type=int, default=3,
                        help='Minimum number of matched peaks (default: 3)')
    annotate_parser.add_argument('--min_spec_usage', type=float, default=0.20,
                        help='Minimum spectral usage (default: 0.20)')

    ####################
    # Subcommand for LC-MS data
    lcms_parser = subparsers.add_parser('lcms', help='Annotate LC-MS spectra in mzML or mzXML format')
    lcms_parser.add_argument('--project_dir', type=str, required=True,
                             help='Path to the project directory')
    lcms_parser.add_argument('--sample_dir', type=str, default='data',
                             help='Directory containing mzML or mzXML files (default: data)')
    lcms_parser.add_argument('--ms1_id_libs', type=str, nargs='*', default=None,
                        help='Optional: One or more paths to MS1 ID library files (.pkl). Paths with spaces should be quoted.')
    lcms_parser.add_argument('--ms2_id_lib', type=str, default=None,
                        help='Optional: Path to MS2 ID library file (.pkl).')

    lcms_parser.add_argument('--parallel', '-p', action='store_true', default=True,
                             help='Run in parallel mode (default: True)')
    lcms_parser.add_argument('--no_parallel', action='store_false', dest='parallel',
                             help='Disable parallel mode')

    lcms_parser.add_argument('--run_rt_correction', action='store_true', default=True,
                             help='Run retention time correction (default: True)')
    lcms_parser.add_argument('--no_run_rt_correction', action='store_false', dest='run_rt_correction',
                             help='Skip retention time correction')

    lcms_parser.add_argument('--run_normalization', action='store_true', default=True,
                             help='Run normalization (default: True)')
    lcms_parser.add_argument('--no_run_normalization', action='store_false', dest='run_normalization',
                             help='Skip normalization')

    lcms_parser.add_argument('--cpu_ratio', type=float, default=0.9,
                        help='CPU usage ratio (default: 0.9)')
    lcms_parser.add_argument('--ms1_tol', type=float, default=0.01,
                        help='MS1 tolerance (default: 0.01)')
    lcms_parser.add_argument('--ms2_tol', type=float, default=0.02,
                        help='MS2 tolerance (default: 0.02)')
    lcms_parser.add_argument('--mass_detect_int_tol', type=float, default=None,
                        help='Mass detection intensity tolerance (default: None, 2e5 for Orbitraps, 5e2 for Q-TOFs, 2e5 for others)')
    lcms_parser.add_argument('--align_mz_tol', type=float, default=0.01,
                        help='Feature alignment m/z tolerance (default: 0.01)')
    lcms_parser.add_argument('--align_rt_tol', type=float, default=0.2,
                        help='Feature alignment retention time tolerance in minutes (default: 0.2)')
    lcms_parser.add_argument('--alignment_drop_by_fill_pct_ratio', type=float, default=0.1,
                        help='Alignment drop by fill percentage ratio (default: 0.1)')
    lcms_parser.add_argument('--peak_cor_rt_tol', type=float, default=0.025,
                        help='Peak-peak correlation retention time tolerance in minutes (default: 0.025)')
    lcms_parser.add_argument('--min_ppc', type=float, default=0.80,
                        help='Minimum peak-peak correlation to form a feature group (default: 0.80)')
    lcms_parser.add_argument('--roi_min_length', type=int, default=5,
                        help='ROI minimum length for a feature (default: 5)')
    lcms_parser.add_argument('--library_search_mztol', type=float, default=0.05,
                        help='Library search m/z tolerance (default: 0.05)')
    lcms_parser.add_argument('--ms1id_score_cutoff', type=float, default=0.7,
                        help='MS1 ID matching score cutoff (default: 0.7)')
    lcms_parser.add_argument('--ms1id_min_matched_peak', type=int, default=3,
                        help='MS1 ID minimum matched peaks (default: 3)')
    lcms_parser.add_argument('--ms1id_min_spec_usage', type=float, default=0.20,
                        help='MS1 ID minimum spectrum usage (default: 0.20)')
    lcms_parser.add_argument('--ms1id_max_prec_rel_int_in_other_ms2', type=float, default=0.01,
                        help='MS1 ID: maximum allowed precursor relative intensity in other MS2 (default: 0.01)')
    lcms_parser.add_argument('--ms2id_score_cutoff', type=float, default=0.7,
                        help='MS2 ID score cutoff (default: 0.7)')
    lcms_parser.add_argument('--ms2id_min_matched_peak', type=int, default=3,
                        help='MS2 ID minimum matched peaks (default: 3)')

    ####################
    # Subcommand for MSI data
    msi_parser = subparsers.add_parser('msi', help='Annotate MS imaging data in imzML & ibd format')
    msi_parser.add_argument('--project_dir', type=str, required=True,
                        help='Project directory containing imzML & ibd files')
    msi_parser.add_argument('--libs', '-l', type=str, nargs='*', default=None,
                        help='One or more paths to library files (.pkl). Paths with spaces should be quoted.')
    msi_parser.add_argument('--n_cores', type=int, default=None,
                        help='Number of cores to use (default: None, use all available cores)')
    msi_parser.add_argument('--mass_detect_int_tol', type=float, default=None,
                        help='Mass detection intensity tolerance (default: None, automatically determined)')
    msi_parser.add_argument('--centroided', action='store_true', default=False,
                        help='Whether the data is centroided (default: False, centroiding will be performed)')
    msi_parser.add_argument('--noise_detection', type=str, default='moving_average',
                        choices=['moving_average', 'percentile'],
                        help='Noise detection method (default: moving_average)')
    msi_parser.add_argument('--sn_factor', type=float, default=3.0,
                        help='Signal-to-noise factor for noise removal (default: 3.0)')
    msi_parser.add_argument('--mz_bin_size', type=float, default=0.01,
                        help='m/z bin size (default: 0.01)')
    msi_parser.add_argument('--min_overlap', type=int, default=10,
                        help='Minimum overlap between ion images (default: 10)')
    msi_parser.add_argument('--min_correlation', type=float, default=0.85,
                        help='Minimum correlation between spectra (default: 0.85)')
    msi_parser.add_argument('--max_cor_depth', type=int, default=1,
                        help='Maximum correlation depth for spatial correlation (default: 1)')
    msi_parser.add_argument('--library_search_mztol', type=float, default=0.05,
                        help='Library search m/z tolerance (default: 0.05)')
    msi_parser.add_argument('--ms1id_score_cutoff', type=float, default=0.7,
                        help='MS1 ID matching score cutoff (default: 0.7)')
    msi_parser.add_argument('--ms1id_min_matched_peak', type=int, default=3,
                        help='MS1 ID minimum matched peaks (default: 3)')
    msi_parser.add_argument('--ms1id_min_spec_usage', type=float, default=0.05,
                        help='MS1 ID minimum spectrum usage (default: 0.05)')

    args = parser.parse_args()

    # Execute appropriate function based on subcommand
    if args.command == 'index':
        index_library(args)
    elif args.command == 'annotate':
        annotate_spectra(args)
    elif args.command == 'lcms':
        run_lcms(args)
    elif args.command == 'msi':
        run_msi(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()


import pandas as pd
import os
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns


def similarity_scores(search='k0'):

    ba_names = pd.read_csv('/data/std_mix/BA_std.tsv', sep='\t')['name'].tolist()
    drug_names = pd.read_csv('/data/std_mix/drug_std.tsv', sep='\t')['name'].tolist()

    data_dir = f'./output_{search}'

    files = [f for f in os.listdir(data_dir) if f.endswith('.tsv') and f.startswith('mixture')]

    all_scores = []
    for f in files:
        df = pd.read_csv(os.path.join(data_dir, f), sep='\t')
        df = df[df['MS1_similarity'].notnull()].reset_index(drop=True)

        # filter out BAs and drugs
        if 'Bile_acids' in f:
            df = df[df['MS1_annotation'].isin(ba_names)].reset_index(drop=True)
        elif 'Drugs' in f:
            df = df[df['MS1_annotation'].isin(drug_names)].reset_index(drop=True)

        df = df[(df['MS1_matched_peak'] >= 4) & (df['MS1_spectral_usage'] >= 0.20)].reset_index(drop=True)

        all_scores += df['MS1_similarity'].tolist()

    print(f'{search}: {len(all_scores)} MS1 scores')
    # save
    np.save(f'data/ms1_scores_{search}.npy', np.array(all_scores))


def plot_score_distributions(searches=['k0', 'k10', 'all'], legend=True, bw_adjust=1.0):
    # Set the font to Arial for all elements
    plt.rcParams['font.family'] = 'sans-serif'
    plt.rcParams['font.sans-serif'] = ['Arial']

    fig, ax = plt.subplots(figsize=(8.5, 5))

    legend_labels = {
        'k0': 'No peak scaling',
        'k10': 'Peak scaling applied',
        'all': 'Search both'
    }
    colors = ['#56648a', '#6280a5', '#8ca5c0', '#8d7e95', '#ca9a96', '#facaa9']
    color_mapping = {
        'k0': colors[4],
        'k10': colors[2],
        'all': colors[0]
    }

    for search in searches:
        scores = np.load(f'data/ms1_scores_{search}.npy')
        if legend:
            sns.kdeplot(scores,
                        label=legend_labels[search],
                        fill=True, ax=ax,
                        color=color_mapping[search],
                        bw_adjust=bw_adjust)
        else:
            sns.kdeplot(scores,
                        fill=True, ax=ax,
                        color=color_mapping[search],
                        bw_adjust=bw_adjust)

    # Add vertical line at score of 0.7
    ax.axvline(x=0.7, color='0.2', linestyle='--', linewidth=1.25)

    # Remove grid
    ax.grid(False)

    # Adjust frame colors
    for spine in ax.spines.values():
        spine.set_edgecolor('0.4')

    # Adjust tick length and padding
    ax.tick_params(axis='x', which='major', length=3, width=1, pad=4, colors='0.3')
    ax.tick_params(axis='y', which='major', length=3, width=1, pad=1.5, colors='0.3')

    ax.set_xlabel('Similarity score', fontname='Arial', color='0.2', fontsize=18, labelpad=10)
    ax.set_ylabel('Density', fontname='Arial', color='0.2', fontsize=18, labelpad=10)

    if legend:
        ax.legend(prop={'family': 'Arial'})

    ax.set_xlim(-0.002, 1.002)

    # Ensure tick labels are also in Arial and colored '0.4'
    for label in ax.get_xticklabels() + ax.get_yticklabels():
        label.set_fontname('Arial')
        label.set_fontsize(12)
        label.set_color('0.3')

    plt.tight_layout()
    plt.savefig('data/ms1_score_distributions.svg', transparent=True)
    plt.show()


if __name__ == '__main__':
    # similarity_scores(search='k0')
    # similarity_scores(search='k10')
    # similarity_scores(search='all')

    plot_score_distributions(legend=False, bw_adjust=0.7)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from umap import UMAP
import seaborn as sns
import pickle


def preprocess_df():
    df = pd.read_csv('all_modes_statistical_analysis.tsv', sep='\t', low_memory=False)
    print(f"Original shape: {df.shape}")

    # remove all cols starting with 'PREF' (pool reference mixture)
    df = df.loc[:, ~df.columns.str.startswith('PREF')]

    # filter out rows with score and matched peaks
    df = df[(df['MS1_similarity'] >= 0.70) & (df['MS1_matched_peak'] >= 4)].reset_index(drop=True)

    # sort by MS1 inchikey, then by MS1_similarity
    df = df.sort_values(['MS1_inchikey', 'MS1_similarity'], ascending=[True, False]).reset_index(drop=True)

    # remove duplicated MS1 inchikey
    df = df.drop_duplicates('MS1_inchikey', keep='first').reset_index(drop=True)

    ########################################################
    # Select data columns
    data_columns = df.columns[49:]
    # Extract the data to plot
    plot_data = df[data_columns].copy()

    # Cap the data at the 95th percentile
    plot_data = plot_data.apply(lambda row: row.clip(upper=np.percentile(row, 95)))

    # log transformation
    plot_data = np.log(plot_data + 1)

    # Perform z-score normalization within each row after log transformation
    plot_data_normalized = plot_data.apply(lambda x: (x - x.mean()) / x.std(), axis=1)

    ########################################################
    meta = pd.read_csv('gnps_df_with_npclassifier_info.tsv', sep='\t', low_memory=False)
    # dict from inchikey to superclass
    inchikey_to_superclass = dict(zip(meta['INCHIKEY'], meta['SUPERCLASS']))
    # dict from inchikey to npsuperclass
    inchikey_to_npsuperclass = dict(zip(meta['INCHIKEY'], meta['npsuperclass']))
    # dict from inchikey to nppathway
    inchikey_to_nppathway = dict(zip(meta['INCHIKEY'], meta['nppathway']))

    # add to df
    df['superclass'] = df['MS1_inchikey'].map(inchikey_to_superclass)
    df['npsuperclass'] = df['MS1_inchikey'].map(inchikey_to_npsuperclass)
    df['nppathway'] = df['MS1_inchikey'].map(inchikey_to_nppathway)

    # save
    df.to_csv('all_modes_statistical_analysis_refined.tsv', sep='\t', index=False)
    pickle.dump(plot_data_normalized, open('plot_data_normalized.pkl', 'wb'))

    return df, plot_data_normalized


def plot_umap(df, plot_data_normalized, color_by='superclass', n_neighbors=15, min_dist=0.1, n_components=2):

    df = df.dropna(subset=[color_by])
    # df = df[~df[color_by].str.contains(';')].reset_index(drop=True)

    df[color_by] = df[color_by].apply(lambda x: x.split(';')[0] if ';' in x else x)

    # Perform UMAP
    umap_model = UMAP(n_neighbors=n_neighbors, min_dist=min_dist, n_components=n_components, random_state=12)
    umap_result = umap_model.fit_transform(plot_data_normalized)

    # Create a DataFrame with UMAP results
    umap_df = pd.DataFrame(data=umap_result, columns=['UMAP1', 'UMAP2'])
    umap_df[color_by] = df[color_by]

    plt.rcParams['font.family'] = 'Arial'

    # Plot

    fig, ax = plt.subplots(figsize=(10, 6))

    ax.spines['left'].set_visible(True)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(True)

    # Remove x and y axis labels
    ax.set_xlabel('UMAP1')
    ax.set_ylabel('UMAP2')

    # Remove x and y axis ticks
    ax.set_xticks([])
    ax.set_yticks([])

    sns.scatterplot(data=umap_df, x='UMAP1', y='UMAP2', hue=color_by, palette='deep', legend=False)
    # plt.title(f'UMAP plot colored by {color_by}')
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0)
    plt.tight_layout()

    # save
    plt.savefig(f'umap.svg', format='svg', bbox_inches='tight', transparent=True)

    plt.show()


from sklearn.manifold import TSNE
import colorsys

def adjust_color(color, saturation_factor=0.7, brightness_factor=1.2):
    """
    Adjust a color by reducing saturation and increasing brightness.
    """
    r, g, b = color[:3]  # Get RGB values, ignore alpha if present
    h, l, s = colorsys.rgb_to_hls(r, g, b)
    s = max(0, min(1, s * saturation_factor))  # Reduce saturation and clip to [0, 1]
    l = max(0, min(1, l * brightness_factor))  # Increase lightness and clip to [0, 1]
    r, g, b = colorsys.hls_to_rgb(h, l, s)
    return (r, g, b)

def plot_tsne(df, plot_data_normalized, fig_size=(10, 6), legend=True,
              color_by='superclass', n_components=2, perplexity=30, n_iter=1000):
    df = df.copy()  # Create a copy to avoid SettingWithCopyWarning
    df = df.dropna(subset=[color_by])

    # Handle multiple values in the color_by column
    df[color_by] = df[color_by].apply(lambda x: x.split(';')[0] if ';' in x else x)

    # Perform t-SNE
    tsne_model = TSNE(n_components=n_components, perplexity=perplexity, max_iter=n_iter, random_state=1)
    tsne_result = tsne_model.fit_transform(plot_data_normalized)

    # Create a DataFrame with t-SNE results
    tsne_df = pd.DataFrame(data=tsne_result, columns=['tSNE1', 'tSNE2'])
    tsne_df[color_by] = df[color_by]

    # Create a custom color palette with 11 distinct colors
    base_colors = plt.cm.tab10(np.linspace(0, 1, 10)).tolist()
    base_colors.append([0.8, 0.8, 0.8, 1])  # Add an 11th color (light gray)

    # Adjust the colors
    colors = [adjust_color(color, saturation_factor=0.7, brightness_factor=1.15) for color in base_colors]

    plt.rcParams['font.family'] = 'Arial'

    # Plot
    fig, ax = plt.subplots(figsize=fig_size)

    ax.spines['left'].set_visible(True)
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(True)

    # Set axis labels
    ax.set_xlabel('')
    ax.set_ylabel('')

    # Remove x and y axis ticks
    ax.set_xticks([])
    ax.set_yticks([])

    # Use the custom palette in the scatterplot
    scatter = sns.scatterplot(data=tsne_df, x='tSNE1', y='tSNE2', hue=color_by, palette=colors, ax=ax)

    # Adjust legend
    if legend:
        plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0)
    else:
        ax.get_legend().remove() if ax.get_legend() else None

    plt.tight_layout()

    # Save the plot
    plt.savefig(f'tsne_{color_by}.svg', format='svg', bbox_inches='tight', transparent=True)

    plt.show()


if __name__ == '__main__':
    # df, plot_data_normalized = preprocess_df()

    ########################################################
    # Load data
    df = pd.read_csv('all_modes_statistical_analysis_refined.tsv', sep='\t', low_memory=False)
    plot_data_normalized = pickle.load(open('plot_data_normalized.pkl', 'rb'))

    # Plot UMAP with different color options
    # plot_umap(df, plot_data_normalized, color_by='superclass', n_neighbors=15, min_dist=0.1)
    # plot_umap(df, plot_data_normalized, color_by='nppathway', n_neighbors=15, min_dist=0.1)

    # Plot t-SNE with different color options
    plot_tsne(df, plot_data_normalized, color_by='superclass', perplexity=30, n_iter=1000,
              fig_size=(10, 6), legend=True)
    # plot_tsne(df, plot_data_normalized, color_by='superclass', perplexity=30, n_iter=1000, fig_size=(6, 6),
    #           legend=False)




import pickle
import pandas as pd
import matplotlib.pyplot as plt

# Load data
ms2_dda_inchikey = pickle.load(open('data/ms2_dda_inchikey.pkl', 'rb'))
ms1_dda_inchikey = pickle.load(open('data/ms1_dda_inchikey.pkl', 'rb'))
ms1_0ev_inchikey = pickle.load(open('data/ms1_0ev_inchikey.pkl', 'rb'))
ms1_10ev_inchikey = pickle.load(open('data/ms1_10ev_inchikey.pkl', 'rb'))
ms1_20ev_inchikey = pickle.load(open('data/ms1_20ev_inchikey.pkl', 'rb'))

metadata_df = pd.read_csv(
    '/Users/shipei/Documents/projects/chemical_conjugate/main/gnps_lib/metadata_df_with_npclassifier_info.tsv',
    sep='\t')

# dict from inchikey to superclass
inchikey_to_superclass = dict(zip(metadata_df['INCHIKEY_14'], metadata_df['SUPERCLASS']))
inchikey_to_class = dict(zip(metadata_df['INCHIKEY_14'], metadata_df['CLASS']))


# Function to count molecules for each superclass in a dataset
def count_molecules(dataset, inchikey_to_class_dict):
    counts = {}
    for inchikey in dataset:
        if inchikey in inchikey_to_class_dict:
            _class = inchikey_to_class_dict[inchikey]
            if pd.notna(_class):  # Only count non-NaN classes
                counts[_class] = counts.get(_class, 0) + 1
    return counts


# Count molecules for each dataset
datasets = {
    'MS1 (20 eV)': ms1_20ev_inchikey,
    'MS1 (10 eV)': ms1_10ev_inchikey,
    'MS1 (0 eV)': ms1_0ev_inchikey,
    'MS1 (DDA)': ms1_dda_inchikey,
    'MS/MS (DDA)': ms2_dda_inchikey
}

all_counts = {name: count_molecules(dataset, inchikey_to_superclass) for name, dataset in datasets.items()}

# Get unique superclasses, excluding NaN values
all_superclasses = list(set.union(*[set(counts.keys()) for counts in all_counts.values()]))

# Create a DataFrame for easier plotting
df = pd.DataFrame(index=datasets.keys(), columns=all_superclasses)
for dataset, counts in all_counts.items():
    for superclass in all_superclasses:
        df.loc[dataset, superclass] = counts.get(superclass, 0)

# Fill NaN values with 0
df = df.fillna(0)

# Calculate total annotations for each class
total_annotations = df.sum()

# Filter classes with total number larger than 10
# filtered_classes = total_annotations[total_annotations > 10].index
# df = df[filtered_classes].reset_index()

# Sort classes by total annotations (descending order)
sorted_classes = df.sum().sort_values(ascending=False).index

# # Sort classes alphabetically
# sorted_classes = sorted(df.columns)

# Reorder DataFrame columns based on sorted superclasses
df = df[sorted_classes]

# Create the plot
plt.rcParams['font.family'] = 'Arial'
fig, ax = plt.subplots(figsize=(6, 3.3))

# Add light grey grid and set it to the background
ax.grid(True, color='0.9', linestyle='-', linewidth=0.4, alpha=0.6, zorder=0)

# Plot circles with individual numbers
for i, dataset in enumerate(df.index):
    for j, superclass in enumerate(df.columns):
        count = df.loc[dataset, superclass]
        if count > 0:
            ax.scatter(j, i, s=3 * count + 50, c='#facaa9', alpha=1.0, edgecolors='none', zorder=2)  # Increase zorder
            ax.text(j, i, str(int(count)), ha='center', va='center', fontsize=8, color='0.15',
                    zorder=3)  # Increase zorder

# Customize the plot
ax.set_xticks(range(len(df.columns)))
ax.set_xticklabels(df.columns, rotation=-35, ha='right', fontsize=8, color='0.15')
ax.xaxis.tick_top()  # Move x-axis labels to the top
ax.xaxis.set_label_position('top')  # Move x-axis label to the top
ax.set_yticks(range(len(datasets)))
ax.set_yticklabels(datasets.keys(), fontsize=10, color='0.15')

# Remove all frames
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)
ax.spines['bottom'].set_visible(False)
ax.spines['left'].set_visible(False)

# Remove ticks
ax.tick_params(axis='both', which='both', length=0)

# Adjust layout
plt.tight_layout()

# Save as SVG
plt.savefig('data/cmpd_class_dot.svg', format='svg', bbox_inches='tight')
plt.show()



import pickle
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import to_rgba
from scipy import stats

# Load data
ms2_inchikeys = set(pickle.load(open('data/ms2_inchikey_gnps.pkl', 'rb')))

# Set Arial font
plt.rcParams['font.family'] = 'sans-serif'
plt.rcParams['font.sans-serif'] = ['Arial']


def plot_combined_distribution(evs=[0, 10, 20], y_range=None, dataset_spacing=1.5):
    plt.figure(figsize=(5, 3))  # Slightly increased height to accommodate legend
    colors = ['#8ca5c0', '#ca9a96']
    whisker_color = '0.5'
    font_color = '0.15'

    all_data = []
    median_positions = []
    p_values = []

    for ev in evs:
        df = pd.read_csv('data/merged_ms1_gnps.tsv', sep='\t', low_memory=False)
        df = df[~df['MS1_inchikey'].isnull()]
        df = df[df['file'].str.contains(f'_{ev}eV')].reset_index(drop=True)
        df = df.sort_values(by='peak_height', ascending=False)
        df['MS1_inchikey14'] = df['MS1_inchikey'].str[:14]
        df = df.drop_duplicates(subset='MS1_inchikey14', keep='first')

        in_ms2 = df[df['MS1_inchikey14'].isin(ms2_inchikeys)]['peak_height']
        not_in_ms2 = df[~df['MS1_inchikey14'].isin(ms2_inchikeys)]['peak_height']

        all_data.extend([in_ms2, not_in_ms2])
        median_positions.append(np.median(in_ms2))

        # Perform Mann-Whitney U test
        statistic, p_value = stats.mannwhitneyu(in_ms2, not_in_ms2, alternative='two-sided')
        p_values.append(p_value)

    positions = []
    for i in range(len(evs)):
        positions.extend([i * dataset_spacing - 0.25, i * dataset_spacing + 0.25])

    bp = plt.boxplot(all_data, positions=positions, widths=0.25, patch_artist=True,
                     showfliers=False, medianprops=dict(color="white", linewidth=1))

    for element in ['whiskers', 'caps']:
        for item in bp[element]:
            item.set_color(whisker_color)

    for i, box in enumerate(bp['boxes']):
        box.set(facecolor=to_rgba(colors[i % 2], alpha=0.85), edgecolor='none')

    plt.yscale('log')
    plt.ylabel('Peak height', fontsize=12, color=font_color)

    plt.xticks([i * dataset_spacing for i in range(len(evs))], [f'MS1 ({ev} eV)' for ev in evs],
               color=font_color, fontsize=12)

    plt.yticks(color=font_color, fontsize=10)

    if y_range is not None:
        plt.ylim(y_range)

    plt.tick_params(axis='x', which='major', length=2.5, width=0.5, color=whisker_color, pad=4)
    plt.tick_params(axis='y', which='major', length=2.5, width=0.5, color=whisker_color, pad=2)
    plt.tick_params(axis='y', which='minor', length=2, width=0.5, color=whisker_color, pad=2)

    # Remove top and right spines
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)

    # Set color for remaining spines
    plt.gca().spines['left'].set_edgecolor(whisker_color)
    plt.gca().spines['bottom'].set_edgecolor(whisker_color)

    for i, data in enumerate(all_data):
        y = data
        x = np.random.normal(positions[i], 0.04, size=len(y))
        plt.scatter(x, y, alpha=0.2, s=2, color=colors[i % 2], zorder=3)

    # Add dots at the median of 'overlap with MS/MS annotation' boxes and connect with a dashed line
    # median_x = [i * dataset_spacing - 0.25 for i in range(len(evs))]
    # plt.plot(median_x, median_positions, color='0.5', linestyle='--', linewidth=1, zorder=4)

    # Add legend
    legend_elements = [
        plt.Rectangle((0, 0), 1, 1, facecolor=colors[0], edgecolor='none', alpha=0.85,
                      label='Overlap with MS/MS annotations'),
        plt.Rectangle((0, 0), 1, 1, facecolor=colors[1], edgecolor='none', alpha=0.85, label='Non-overlap'),
    ]
    plt.legend(handles=legend_elements, loc='upper center', bbox_to_anchor=(0.5, 1.1),
               ncol=2, fontsize=9, frameon=False)

    # Add p-values to the plot
    for i, p_value in enumerate(p_values):
        plt.text(i * dataset_spacing, plt.ylim()[1] - 1e9, f'P = {p_value:.2e}',
                 horizontalalignment='center', verticalalignment='bottom',
                 fontsize=9, color=font_color)

    plt.tight_layout()
    plt.savefig('data/combined_peak_height_distribution.svg', transparent=True, bbox_inches='tight')
    plt.show()
    plt.close()


y_range = (1.5e5, 1.5e9)
plot_combined_distribution(evs=[0, 10, 20], y_range=y_range, dataset_spacing=1.5)


import pandas as pd
import os


def merge_results():

    all_results = pd.DataFrame()

    # Walk through the directory
    for root, dirs, files in os.walk('/imaging/MTBLS313'):
        for file in files:
            # Check if the file ends with .tsv
            if file.endswith('_all.tsv'):
                # Construct the full file path
                full_path = os.path.join(root, file)

                # get its parent directory
                parent_dir = os.path.basename(os.path.dirname(full_path))

                # Read the file
                try:
                    result_df = pd.read_csv(full_path, sep='\t', low_memory=False)
                    result_df['sample'] = parent_dir
                except:
                    continue

                all_results = pd.concat([all_results, result_df])

    all_results.to_csv('all_brain_annotations.tsv', sep='\t', index=False)


if __name__ == '__main__':
    merge_results()




import pickle
import matplotlib.pyplot as plt
from matplotlib_venn import venn3

ms2_inchikeys = set(pickle.load(open('../data/ms2_inchikey_gnps.pkl', 'rb')))
ms1_0ev_inchikeys = set(pickle.load(open('../data/ms1_inchikey_gnps_0ev.pkl', 'rb')))
ms1_10ev_inchikeys = set(pickle.load(open('../data/ms1_inchikey_gnps_10ev.pkl', 'rb')))
# ms1_20ev_inchikeys = set(pickle.load(open('data/ms1_inchikey_gnps_20ev.pkl', 'rb')))


colors = ['#56648a', '#6280a5', '#8ca5c0', '#8d7e95', '#ca9a96', '#facaa9']

# Set the font to Arial
plt.rcParams['font.family'] = 'Arial'

# Create the figure and axis
fig, ax = plt.subplots(figsize=(1.6, 1.6))

# Create the Venn diagram
v = venn3([ms2_inchikeys, ms1_0ev_inchikeys, ms1_10ev_inchikeys],
          # set_labels=('MS/MS', 'MS1 (0 eV)', 'MS1 (10 eV)'),
          set_labels=('', '', ''),
          set_colors=(colors[0], colors[3], colors[5]))

# Customize the diagram
for text in v.set_labels:
    text.set_fontsize(8)
for text in v.subset_labels:
    text.set_fontsize(6)

# Add circles for a cleaner look
# venn3_circles([ms2_inchikeys, ms1_0ev_inchikeys, ms1_10ev_inchikeys], linewidth=0.5, color="gray")

plt.tight_layout()

# Save the figure (optional)
plt.savefig('ms_venn_diagram_gnps.svg', transparent=True)

# Show the plot
plt.show()


import numpy as np
import pyimzml.ImzMLParser as imzml
import matplotlib.pyplot as plt
from scipy.ndimage import median_filter
import matplotlib.colors as colors


def get_ion_image_from_imzml(imzml_file, target_mz, tolerance=0.01, intensity_threshold=0):
    parser = imzml.ImzMLParser(imzml_file)

    # Get the maximum x and y coordinates from all spectra
    max_x = max(x for x, y, z in parser.coordinates)
    max_y = max(y for x, y, z in parser.coordinates)

    # Initialize the image array with the maximum dimensions
    image = np.zeros((max_y + 1, max_x + 1))

    for idx, (x, y, z) in enumerate(parser.coordinates):
        mzs, ints = parser.getspectrum(idx)
        mask = np.abs(mzs - target_mz) <= tolerance
        matched_intensities = ints[mask]

        if len(matched_intensities) > 0:
            intensity = np.max(matched_intensities)
            if intensity > intensity_threshold:
                image[y, x] = intensity

    return image, parser.coordinates


def plot_ion_image(image, title=None, colormap='viridis',
                   x_range=None, y_range=None,
                   norm_max_intensity=3e7,
                   hot_spot_removal=False, hot_spot_percentile=99.9,
                   median_filter=False,
                   save=False, name='ion_image.svg'):
    """
    Plot the ion image, with options to plot partial data.

    Parameters:
    - image: 2D numpy array representing the ion image
    - title: Optional title for the plot
    - colormap: Colormap to use for the image (default: 'viridis')
    - x_range: Tuple of (start, end) for x-axis. If None, plot all. (default: None)
    - y_range: Tuple of (start, end) for y-axis. If None, plot all. (default: None)
    """
    # Apply ranges if specified
    if x_range is not None:
        image = image[:, x_range[0]:x_range[1]]
    if y_range is not None:
        image = image[y_range[0]:y_range[1], :]

    # Apply processing
    if hot_spot_removal:
        percentile_top = np.percentile(image, hot_spot_percentile)
        image[image > percentile_top] = percentile_top

    if median_filter:
        image = median_filter(image, size=3)

    plt.figure(figsize=(10, 8))

    # Create a custom colormap using the later 90% of viridis
    original_cmap = plt.get_cmap(colormap)
    colors_array = original_cmap(np.linspace(0.10, 1, 256))
    custom_cmap = colors.LinearSegmentedColormap.from_list("custom_" + colormap, colors_array)

    plt.imshow(image, cmap=custom_cmap, vmin=0, vmax=norm_max_intensity)
    plt.colorbar(label='Intensity')

    if title:
        plt.title(title)

    plt.axis('off')
    plt.tight_layout()

    if save:
        plt.savefig(name, transparent=True)
    plt.show()


from scipy.stats import pearsonr
def calculate_ion_image_correlation(image1, image2):
    """
    Calculate the intensity correlation between two ion images,
    considering only pixels where both images have intensities > 0.

    Parameters:
    - image1: 2D numpy array representing the first ion image
    - image2: 2D numpy array representing the second ion image

    Returns:
    - correlation_coefficient: Pearson correlation coefficient
    - p_value: Two-tailed p-value
    """
    # Ensure the images have the same shape
    if image1.shape != image2.shape:
        raise ValueError("The two images must have the same shape")

    # Flatten the 2D images into 1D arrays
    flat_image1 = image1.flatten()
    flat_image2 = image2.flatten()

    # Create a mask for pixels where both images have intensity > 0
    mask = (flat_image1 > 0) & (flat_image2 > 0)

    # Apply the mask to both flattened images
    valid_intensities1 = flat_image1[mask]
    valid_intensities2 = flat_image2[mask]

    # Check if there are enough valid pixels for correlation
    if len(valid_intensities1) < 2:
        return None, None  # Not enough data points for correlation

    # Calculate Pearson correlation
    correlation_coefficient, p_value = pearsonr(valid_intensities1, valid_intensities2)

    return correlation_coefficient, p_value


if __name__ == "__main__":

    ##########################################
    # 808.585116
    t_mz = 612.556176 #612.556176
    imzml_file = "/imaging/hepatocytes/20171107_F4_DHBpos_p70_s50.imzML"
    ion_image, _ = get_ion_image_from_imzml(imzml_file, t_mz, tolerance=t_mz * 5e-6, intensity_threshold=0)

    max_intensity = np.max(ion_image) * 0.4
    colormap = 'viridis'  # viridis, cividis, magma, inferno, plasma
    # Plot partial image
    plot_ion_image(ion_image,
                   title=None, colormap=colormap,
                   x_range=(1, None), y_range=(1, None),
                   norm_max_intensity=max_intensity,
                   hot_spot_removal=False, hot_spot_percentile=99,
                   median_filter=False,
                   save=True, name='DG_16_0_18_1_ion_image.svg')





__version__ = "0.1.1"
__author__ = "Shipei Xing"
__email__ = "philipxsp@hotmail.com"
__author_email__ = "philipxsp@hotmail.com"

__copyright__ = "Copyright 2024 Shipei Xing"
__license_expression__ = "Apache 2.0"

__url__ = "https://github.com/Philipbear/ms1_id"

__title__ = "ms1_id"


import multiprocessing as mp
import os
import pickle

import numpy as np
from scipy.sparse import csr_matrix
from tqdm import tqdm

from ms1_id.msi.utils_imaging import PseudoMS2
from ms1_id.msi.export_msi import write_pseudoms2_to_mgf


def generate_pseudo_ms2(mz_values, intensity_matrix, correlation_matrix,
                        n_processes=None, min_cluster_size=6,
                        min_cor=0.90, max_cor_depth=1,
                        save=False, save_dir=None, chunk_size=1000):
    """
    Generate pseudo MS2 spectra for imaging data using chunked parallel processing
    """
    # Check if result files exist
    if save_dir:
        save_path = os.path.join(save_dir, 'pseudo_ms2_spectra.pkl')
        if os.path.exists(save_path):
            print("Loading existing pseudo MS2 spectra...")
            with open(save_path, 'rb') as f:
                return pickle.load(f)

    # Perform clustering
    pseudo_ms2_spectra = _perform_clustering(mz_values, correlation_matrix,
                                             n_processes=n_processes,
                                             min_cor=min_cor,
                                             max_cor_depth=max_cor_depth,
                                             min_cluster_size=min_cluster_size,
                                             chunk_size=chunk_size)

    # Assign intensity values
    _assign_intensities(pseudo_ms2_spectra, intensity_matrix)

    if save and save_dir:
        pkl_path = os.path.join(save_dir, 'pseudo_ms2_spectra.pkl')
        with open(pkl_path, 'wb') as f:
            pickle.dump(pseudo_ms2_spectra, f)

        mgf_path = os.path.join(save_dir, 'pseudo_ms2_spectra.mgf')
        write_pseudoms2_to_mgf(pseudo_ms2_spectra, mgf_path)

    return pseudo_ms2_spectra


def _process_chunk(args):
    """
    Process a chunk of m/z values for clustering, considering max_cor_depth and correlation threshold.
    """
    start_idx, end_idx, mz_values, correlation_matrix, min_cor, min_cluster_size, max_cor_depth = args
    chunk_results = []

    def find_correlated_mzs(target_idx, current_depth, visited):
        if current_depth > max_cor_depth:
            return set()

        row = correlation_matrix[target_idx].toarray().flatten()

        if current_depth == 0:
            # correlated_indices = set(np.where((mz_values <= mz_values[target_idx] + 1e-2) & (row > 0))[0]) - visited
            correlated_indices = set(np.where(row > 0)[0]) - visited
        else:
            # For depth >= 1, apply the correlation threshold with the original target
            original_target_row = correlation_matrix[start_idx].toarray().flatten()
            # correlated_indices = set(np.where((mz_values <= mz_values[target_idx] + 1e-2) &
            #                                   (row > 0) & (original_target_row >= (min_cor - 0.10)))[0]) - visited
            correlated_indices = set(np.where((row > 0) & (original_target_row >= (min_cor - 0.10)))[0]) - visited

        if current_depth < max_cor_depth:
            for idx in correlated_indices.copy():
                correlated_indices.update(find_correlated_mzs(idx, current_depth + 1, visited | correlated_indices))

        return correlated_indices

    for i in range(start_idx, end_idx):
        mz = mz_values[i]
        correlated_indices = find_correlated_mzs(i, 0, set())

        if len(correlated_indices) >= min_cluster_size:
            cluster_mzs = mz_values[list(correlated_indices)]
            indices = sorted(correlated_indices)
            chunk_results.append(PseudoMS2(mz, i, cluster_mzs.tolist(), [0] * len(cluster_mzs), indices))

    return chunk_results


def _perform_clustering(mz_values, correlation_matrix, n_processes=None, min_cor=0.90,
                        min_cluster_size=6, max_cor_depth=1, chunk_size=800):
    """
    Perform clustering on m/z values based on correlation scores using chunked multiprocessing.
    """
    if not isinstance(correlation_matrix, csr_matrix):
        correlation_matrix = csr_matrix(correlation_matrix)

    n_processes = n_processes or mp.cpu_count()

    # Prepare chunks
    n_chunks = (len(mz_values) + chunk_size - 1) // chunk_size
    chunks = [(i * chunk_size, min((i + 1) * chunk_size, len(mz_values)),
               mz_values, correlation_matrix, min_cor, min_cluster_size, max_cor_depth)
              for i in range(n_chunks)]

    # Process chunks in parallel
    with mp.Pool(processes=n_processes) as pool:
        results = list(tqdm(pool.imap(_process_chunk, chunks),
                            total=len(chunks), desc="Processing chunks"))

    # Flatten results
    pseudo_ms2_spectra = [spectrum for chunk_result in results for spectrum in chunk_result]

    return pseudo_ms2_spectra


def _assign_intensities(pseudo_ms2_spectra, intensity_matrix):
    """
    Assign intensity values to pseudo MS2 spectra.
    """
    for spectrum in tqdm(pseudo_ms2_spectra, desc="Assigning intensities"):
        # Get the intensities for all m/z values in this PseudoMS1 object
        intensities = intensity_matrix[spectrum.indices, :]

        # Get the intensities for the target m/z across all spectra
        t_mz_intensities = intensity_matrix[spectrum.t_mz_idx, :]

        # Find the spectrum with the highest intensity at the target m/z
        max_spectrum_index = np.argmax(t_mz_intensities)

        spectrum.intensities = intensities[:, max_spectrum_index].tolist()



import requests
import json


def get_inchikey_from_refmet_name(refmet_name):
    base_url = "https://www.metabolomicsworkbench.org/rest/refmet/name/"
    url = f"{base_url}{refmet_name}/inchi_key"

    try:
        response = requests.get(url)
        data = json.loads(response.text)
        return data['inchi_key'] if data else None
    except:
        return None

# Example usage:
# inchikey = get_inchikey_from_refmet_name("Cholesterol")
# print(inchikey)


def get_refmet_name_from_inchikey(inchikey):
    base_url = "https://www.metabolomicsworkbench.org/rest/refmet/inchi_key/"
    url = f"{base_url}{inchikey}/name"

    try:
        response = requests.get(url)
        data = json.loads(response.text)
        return data['name'] if data else None
    except:
        return None

# Example usage:
# refmet_name = get_refmet_name_from_inchikey("HVYWMOMLDIMFJA-DPAQBDIFSA-N")
# print(refmet_name)


def get_all_info_from_inchikey(inchikey):
    base_url = "https://www.metabolomicsworkbench.org/rest/compound/inchi_key/"
    url = f"{base_url}{inchikey}/all"

    try:
        response = requests.get(url)
        data = json.loads(response.text)
        return data if data else None
    except:
        return None

# Example usage:
# all_info = get_all_info_from_inchikey("HVYWMOMLDIMFJA-DPAQBDIFSA-N")
# print(json.dumps(all_info, indent=2))


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats


def get_peak_height_ls():
    df = pd.read_pickle('data/aligned_feature_table_all.pkl')

    # max intensity
    df['max_intensity'] = df.iloc[:, 22:28].max(axis=1)

    ms2_dda_ph_ls = df[df['similarity'].notnull()]['max_intensity'].tolist()
    ms1_dda_ph_ls = df[df['ms1_dda_inchikeys'].notnull()]['max_intensity'].tolist()
    ms1_0ev_ph_ls = df[df['ms1_0ev_inchikeys'].notnull()]['max_intensity'].tolist()
    ms1_10ev_ph_ls = df[df['ms1_10ev_inchikeys'].notnull()]['max_intensity'].tolist()
    ms1_20ev_ph_ls = df[df['ms1_20ev_inchikeys'].notnull()]['max_intensity'].tolist()

    return ms2_dda_ph_ls, ms1_dda_ph_ls, ms1_0ev_ph_ls, ms1_10ev_ph_ls, ms1_20ev_ph_ls


# Set up the plot
plt.rcParams['font.family'] = 'Arial'
fig, ax = plt.subplots(figsize=(5, 1.85))

# Get peak height data
peak_heights = get_peak_height_ls()

# Labels for x-axis
labels = ['MS/MS\n(DDA)', 'MS1\n(DDA)', 'MS1\n(0 eV)', 'MS1\n(10 eV)', 'MS1\n(20 eV)']

# Plot scatter points behind the box
for i, heights in enumerate(peak_heights):
    x = np.random.normal(i + 1, 0.04, len(heights))
    ax.scatter(x, heights, alpha=0.1, s=2, color='blue', edgecolors='none', zorder=1)

# Create box plot
box_color = 'lightgrey'
box_width = 0.3  # Adjust this value to change the width of the boxes
box_plot = ax.boxplot(peak_heights, patch_artist=True, tick_labels=labels, showfliers=False, widths=box_width)

# Customize box appearance
for element in ['boxes', 'fliers', 'means', 'medians']:
    plt.setp(box_plot[element], color='none')

# Set only whiskers color
plt.setp(box_plot['whiskers'], color='0.5')
plt.setp(box_plot['caps'], color='0.5')

for median in box_plot['medians']:
    median.set(color='white', linewidth=1.5)

for patch in box_plot['boxes']:
    patch.set_facecolor(box_color)
    patch.set_edgecolor('none')
    patch.set_alpha(0.8)

# Customize the plot
ax.set_ylabel('Peak height', fontsize=12, fontname='Arial', labelpad=8, color='0.2')
ax.set_yscale('log')

# Set y-axis limits and show minor ticks
ax.set_ylim(2e5, 1.5e8)
ax.yaxis.set_minor_locator(plt.LogLocator(subs=np.arange(2, 10)))

# Customize ticks and remove top and right spines
ax.tick_params(axis='x', which='major', labelsize=10, pad=5, length=2, colors='0.2')
ax.tick_params(axis='y', which='major', labelsize=8, pad=2, length=2, colors='0.2')
ax.tick_params(axis='y', which='minor', length=1, colors='0.2')  # Show minor ticks
ax.spines['top'].set_visible(False)
ax.spines['right'].set_visible(False)
ax.spines['bottom'].set_color('0.5')
ax.spines['left'].set_color('0.5')

# Set x-axis and y-axis label font to Arial and color to 0.2
for label in ax.get_xticklabels() + ax.get_yticklabels():
    label.set_fontname('Arial')
    label.set_color('0.2')

# Adjust the space between the first box and y-axis
plt.xlim(0.3, len(labels) + 0.5)

# # Calculate Mann-Whitney U test and add p-values with lines
# ms2_dda = peak_heights[0]
# y_max = ax.get_ylim()[1]
# line_heights = [0.45, 0.8, 1.15, 1.5]  # Adjusted heights for each comparison

# for i, (other_data, line_height) in enumerate(zip(peak_heights[1:], line_heights), start=1):
#     statistic, p_value = stats.mannwhitneyu(ms2_dda, other_data, alternative='two-sided')
#     p_text = f'$\it{{P}}$ = {p_value:.2e}'
#
#     # Draw line
#     x1, x2 = 1, i + 1
#     y = y_max * line_height
#     ax.plot([x1, x1, x2, x2], [y, y + y_max * 0.02, y + y_max * 0.02, y],
#             lw=1, color='0.5')
#
#     # Add p-value text
#     ax.text((x1 + x2) * 0.5, y, p_text,
#             ha='center', va='bottom', fontsize=8, color='0.2')
#
# # Adjust y-axis limit to show the lines and p-values
# ax.set_ylim(2e5, y_max * 1.9)

print("Cap coordinates for each box:")
for i in range(len(labels)):
    bottom_cap = box_plot['caps'][2*i].get_ydata()[0]
    top_cap = box_plot['caps'][2*i + 1].get_ydata()[0]
    print(f"{labels[i]}:")
    print(f"  Bottom cap: {bottom_cap}")
    print(f"  Top cap: {top_cap}")

# # Add dashed lines connecting top and bottom caps
# bright_color = '#FF4500'  # Bright orange-red color
# for i in range(2, 4):  # Connect 3rd to 4th, then 4th to 5th
#     # Connect top caps
#     top_cap_y = box_plot['caps'][2 * i - 1].get_ydata()[0]  # Y-coordinate of current box's top cap
#     next_top_cap_y = box_plot['caps'][2 * i + 1].get_ydata()[0]  # Y-coordinate of next box's top cap
#     ax.plot([i + 1, i + 2], [top_cap_y, next_top_cap_y], color=bright_color, linestyle='--', linewidth=1, zorder=3)
#
#     # Connect bottom caps
#     bottom_cap_y = box_plot['caps'][2 * i - 2].get_ydata()[0]  # Y-coordinate of current box's bottom cap
#     next_bottom_cap_y = box_plot['caps'][2 * i].get_ydata()[0]  # Y-coordinate of next box's bottom cap
#     ax.plot([i + 1, i + 2], [bottom_cap_y, next_bottom_cap_y], color=bright_color, linestyle='--', linewidth=1,
#             zorder=3)

plt.tight_layout()

# Save as svg
plt.savefig('data/peak_height_boxplot.svg', transparent=True, bbox_inches='tight')

plt.show()


"""
create a workflow for MS1_ID using masscube backend.
for single file mode.
"""

import os

from masscube.feature_grouping import annotate_isotope
from masscube.params import Params, find_ms_info
from masscube.raw_data_utils import MSData

from ms1_id.lcms.annotate_adduct import annotate_adduct
from ms1_id.lcms.annotate_ms2 import annotate_rois
from ms1_id.lcms.calculate_ppc import calc_all_ppc
from ms1_id.lcms.export import write_single_file
from ms1_id.lcms.group_ppc import generate_pseudo_ms2
from ms1_id.lcms.reverse_matching import ms1_id_annotation

# default parameters
orbitrap_mass_detect_int_tol = 20000
tof_mass_detect_int_tol = 500


def main_workflow_single(file_path,
                         ms1id_library_path=None, ms2id_library_path=None,
                         ms1_id=True, ms2_id=False,
                         ms1_tol=0.01, ms2_tol=0.015, mass_detect_int_tol=None,
                         peak_cor_rt_tol=0.015,
                         min_ppc=0.8, roi_min_length=3,
                         library_search_mztol=0.05,
                         ms1id_score_cutoff=0.7, ms1id_min_matched_peak=6, ms1id_min_spec_usage=0.1,
                         ms1id_max_prec_rel_int_in_other_ms2=0.01,
                         ms2id_score_cutoff=0.7, ms2id_min_matched_peak=6,
                         plot_bpc=False, out_dir=None):
    """
    Untargeted feature detection from a single file (.mzML or .mzXML).
    """

    ms_type, ion_mode, centroid = find_ms_info(file_path)

    if not centroid:
        raise ValueError("The file is not centroided.")

    # init a new config object
    config = init_config_single(ms_type, ion_mode, ms1id_library_path, ms2id_library_path,
                                mz_tol_ms1=ms1_tol, mz_tol_ms2=ms2_tol,
                                mass_detect_int_tol=mass_detect_int_tol)

    # create a MSData object
    d = MSData()

    # read raw data
    print('Reading raw data...')
    d.read_raw_data(file_path, config)

    # detect region of interests (ROIs)
    print('Detecting ROIs...')
    d.find_rois()

    # cut ROIs
    d.cut_rois()

    # label short ROIs, find the best MS2, and sort ROIs by m/z
    print('Summarizing ROIs...')
    d.summarize_roi()

    # annotate isotopes, adducts
    print('Annotating isotopes and adducts...')
    annotate_isotope(d)
    annotate_adduct(d)

    # generate pseudo ms2 spec for ms1_id
    pseudo_ms2_spectra = []
    if ms1_id and config.ms1id_library_path is not None:
        print("MS1 ID annotation...")

        # calc peak-peak correlations for feature groups and output
        print('Calculating peak-peak correlations...')
        ppc_matrix = calc_all_ppc(d, rt_tol=peak_cor_rt_tol, roi_min_length=roi_min_length, min_ppc=min_ppc,
                                  save=False)

        # generate pseudo ms2 spec, for ms1_id
        print('Generating pseudo MS2 spectra...')
        pseudo_ms2_spectra = generate_pseudo_ms2(d, ppc_matrix,
                                                 mz_tol=ms1_tol,
                                                 min_ppc=min_ppc,
                                                 min_cluster_size=ms1id_min_matched_peak,
                                                 roi_min_length=roi_min_length,
                                                 save_dir=os.path.dirname(file_path))

        # perform rev cos search
        print('Performing MS1 ID annotation...')
        pseudo_ms2_spectra = ms1_id_annotation(pseudo_ms2_spectra, config.ms1id_library_path,
                                               mz_tol=library_search_mztol,
                                               ion_mode=ion_mode,
                                               max_prec_rel_int_in_other_ms2=ms1id_max_prec_rel_int_in_other_ms2,
                                               score_cutoff=ms1id_score_cutoff,
                                               min_matched_peak=ms1id_min_matched_peak,
                                               min_spec_usage=ms1id_min_spec_usage)

        # # write out raw ms1 id results
        # write_ms1_id_results(pseudo_ms2_spectra, out_dir=os.path.dirname(file_path))

    # annotate MS2 spectra
    if ms2_id and config.ms2id_library_path is not None:
        print("Annotating MS2 spectra...")
        annotate_rois(d, ms2id_score_cutoff=ms2id_score_cutoff,
                      ms2id_min_matched_peak=ms2id_min_matched_peak,
                      ion_mode=ion_mode)

    if plot_bpc:
        bpc_path = os.path.splitext(file_path)[0] + "_bpc.png"
        d.plot_bpc(label_name=True, output_dir=bpc_path)

    # output single file to a tsv file, in the same directory as the raw file
    if out_dir is None:
        out_dir = os.path.dirname(file_path)
    out_path = os.path.join(out_dir, os.path.splitext(os.path.basename(file_path))[0] + "_feature_table.tsv")
    write_single_file(d, pseudo_ms2_spectra, out_path)

    return d


def init_config_single(ms_type, ion_mode, ms1id_library_path, ms2id_library_path,
                       mz_tol_ms1=0.01, mz_tol_ms2=0.015,
                       mass_detect_int_tol=None):
    # init
    config = Params()

    if mass_detect_int_tol is not None:
        config.int_tol = mass_detect_int_tol
    else:
        if ms_type == "orbitrap":
            config.int_tol = orbitrap_mass_detect_int_tol
        elif ms_type == "tof":
            config.int_tol = tof_mass_detect_int_tol
        else:
            config.int_tol = 20000

    ##########################
    # The project
    config.project_dir = None  # Project directory, character string
    config.sample_names = None  # Absolute paths to the raw files, without extension, list of character strings
    config.sample_groups = None  # Sample groups, list of character strings
    config.sample_group_num = None  # Number of sample groups, integer
    config.sample_dir = None  # Directory for the sample information, character string
    config.single_file_dir = None  # Directory for the single file output, character string
    config.annotation_dir = None  # Directory for the annotation output, character string
    config.chromatogram_dir = None  # Directory for the chromatogram output, character string
    # config.network_dir = None             # Directory for the network output, character string
    config.statistics_dir = None  # Directory for the statistical analysis output, character string

    # MS data acquisition
    config.rt_range = [0.0, 1000.0]  # RT range in minutes, list of two floats
    config.ion_mode = ion_mode  # Ionization mode, "positive" or "negative", character string

    # Feature detection
    config.mz_tol_ms1 = mz_tol_ms1  # m/z tolerance for MS1, default is 0.01
    config.mz_tol_ms2 = mz_tol_ms2  # m/z tolerance for MS2, default is 0.015
    # config.int_tol = 30000  # Intensity tolerance, default is 30000 for Orbitrap and 1000 for other instruments, integer
    config.roi_gap = 30  # Gap within a feature, default is 30 (i.e. 30 consecutive scans without signal), integer
    config.ppr = 0.8  # Peak-peak correlation threshold for feature grouping, default is 0.7

    # Parameters for feature alignment
    config.align_mz_tol = 0.01  # m/z tolerance for MS1, default is 0.01
    config.align_rt_tol = 0.2  # RT tolerance, default is 0.2
    config.run_rt_correction = True  # Whether to perform RT correction, default is True
    config.min_scan_num_for_alignment = 5  # Minimum scan number a feature to be aligned, default is 6

    # Parameters for feature annotation
    config.ms1id_library_path = ms1id_library_path  # Path to the MS1 library (.pkl), character string
    config.ms2id_library_path = ms2id_library_path  # Path to the MS/MS library (.pkl), character string
    config.ms2_sim_tol = 0.7  # MS2 similarity tolerance, default is 0.7

    # Parameters for normalization
    config.run_normalization = False  # Whether to normalize the data, default is False
    config.normalization_method = "pqn"  # Normalization method, default is "pqn" (probabilistic quotient normalization), character string

    # Parameters for output
    config.output_single_file = False  # Whether to output the processed individual files to a csv file
    config.output_aligned_file = False  # Whether to output aligned features to a csv file

    # Statistical analysis
    config.run_statistics = False  # Whether to perform statistical analysis

    # # Network analysis
    # config.run_network = False            # Whether to perform network analysis

    # Visualization
    config.plot_bpc = False  # Whether to plot base peak chromatogram
    config.plot_ms2 = False  # Whether to plot mirror plots for MS2 matching

    return config




import numpy as np
import pickle
import matplotlib.pyplot as plt
import os
from numba import njit
from pyimzml.ImzMLParser import ImzMLParser


def get_spectrum_from_imzml(imzml_file, coordinate=None, index=None):
    """
    Get a spectrum from an imzML file by coordinate or index.
    Prints out the coordinate range of the imzML file.

    Parameters:
    - imzml_file (str): Path to the imzML file.
    - coordinate (tuple): (x, y) coordinate of the spectrum to retrieve.
    - index (int): Index of the spectrum to retrieve.

    Returns:
    - mzs (np.array): m/z values of the spectrum.
    - intensities (np.array): Intensity values of the spectrum.

    Note: Provide either coordinate or index, not both.
    """
    if coordinate is None and index is None:
        raise ValueError("Must provide either coordinate or index.")
    if coordinate is not None and index is not None:
        raise ValueError("Provide either coordinate or index, not both.")

    # Parse the imzML file
    imzml = ImzMLParser(imzml_file)

    # Get coordinate range
    x_coords, y_coords, _ = zip(*imzml.coordinates)
    x_min, x_max = min(x_coords), max(x_coords)
    y_min, y_max = min(y_coords), max(y_coords)
    print(f"Coordinate range: X: {x_min} to {x_max}, Y: {y_min} to {y_max}")
    print(f"Total number of spectra: {len(imzml.coordinates)}")

    if coordinate is not None:
        x, y = coordinate
        # Find the index of the spectrum at the given coordinate
        for idx, (x_, y_, _) in enumerate(imzml.coordinates):
            if x_ == x and y_ == y:
                index = idx
                break
        else:
            raise ValueError(f"No spectrum found at coordinate ({x}, {y})")

    # Get the spectrum at the specified index
    mzs, intensities = imzml.getspectrum(index)

    return np.array(mzs), np.array(intensities)


@njit
def apply_moving_average_numba(mz_array, intensity_array, mz_window=50.0,
                               percentage_lowest=0.05, factor=5):
    """
    Apply moving average algorithm to a single mass spectrum using an m/z-based window.
    This function is optimized with Numba.

    :param mz_array: numpy array of m/z values
    :param intensity_array: numpy array of intensity values
    :param mz_window: size of the moving window in Da (default: 100.0)
    :param n_lowest: number of lowest points to consider in each window (default: 5)
    :param factor: factor to multiply the mean of the lowest points (default: 5.0)
    :return: tuple of (filtered_mz_array, filtered_intensity_array)
    """
    filtered_mz = []
    filtered_intensity = []
    noise_mz = []
    noise_intensity = []

    for i in range(len(mz_array)):
        mz_min = mz_array[i] - mz_window / 2
        mz_max = mz_array[i] + mz_window / 2

        window_intensities = intensity_array[(mz_array >= mz_min) & (mz_array <= mz_max)]

        if len(window_intensities) > 0:
            positive_intensities = window_intensities[window_intensities > 0]
            if len(positive_intensities) > 0:
                sorted_intensities = np.sort(positive_intensities)
                n_lowest = max(1, int(len(sorted_intensities) * percentage_lowest))
                lowest_n = sorted_intensities[:n_lowest]
                baseline = factor * np.mean(lowest_n)
            else:
                baseline = 0.0
        else:
            baseline = 0.0

        if intensity_array[i] > baseline:
            filtered_mz.append(mz_array[i])
            filtered_intensity.append(intensity_array[i])
        else:
            noise_mz.append(mz_array[i])
            noise_intensity.append(intensity_array[i])

    return np.array(filtered_mz), np.array(filtered_intensity), np.array(noise_mz), np.array(noise_intensity)


def load_ms_imaging_data(data_dir):
    """
    Load pre-generated MS imaging data from files.

    Parameters:
    - data_dir: string, path to the directory containing the data files

    Returns:
    - mz_values: numpy array of m/z values
    - intensity_matrix: 2D numpy array where each row corresponds to an m/z value and each column to a pixel
    - coordinates: list of (x, y) coordinates for each pixel
    """
    mz_values_path = os.path.join(data_dir, 'mz_values.npy')
    intensity_matrix_path = os.path.join(data_dir, 'intensity_matrix.npy')
    coordinates_path = os.path.join(data_dir, 'coordinates.pkl')

    # Load mz_values
    mz_values = np.load(mz_values_path)

    # Load intensity_matrix
    intensity_matrix = np.load(intensity_matrix_path)

    # Load coordinates
    with open(coordinates_path, 'rb') as f:
        coordinates = pickle.load(f)

    return mz_values, intensity_matrix, coordinates


def get_ms_spectrum(mz_values, intensity_matrix, pixel_index):
    """
    Get MS spectrum for a given pixel.

    Parameters:
    - mz_values: numpy array of m/z values
    - intensity_matrix: 2D numpy array where each row corresponds to an m/z value and each column to a pixel
    - pixel_index: index of the pixel to retrieve

    Returns:
    - tuple of (mz_array, intensity_array)
    """
    return mz_values, intensity_matrix[:, pixel_index]


def plot_ms_spectrum(mzs, intensities, title=None,
                     x_range=None, y_range=None):
    """
    Plot MS spectrum for a given pixel.

    Parameters:
    - mz_values: numpy array of m/z values
    - intensity_matrix: 2D numpy array where each row corresponds to an m/z value and each column to a pixel
    - coordinates: list of (x, y) coordinates for each pixel
    - pixel_index: index of the pixel to plot
    - title: optional title for the plot
    - xlim: optional tuple of (min, max) for x-axis limits
    - ylim: optional tuple of (min, max) for y-axis limits

    Returns:
    - fig, ax: matplotlib figure and axis objects
    """

    # Create the plot
    fig, ax = plt.subplots(figsize=(15, 6))

    # Plot the spectrum
    ax.vlines(mzs, [0], intensities, linewidth=1)

    # Set labels and title
    ax.set_xlabel('m/z')
    ax.set_ylabel('Intensity')

    # Set axis limits if provided
    if x_range:
        ax.set_xlim(x_range[0], x_range[1])
    if y_range:
        ax.set_ylim(y_range[0], y_range[1])

    # Add grid
    ax.grid(True, linestyle='--', alpha=0.7)

    # Add title
    if title:
        ax.set_title(title)

    # Adjust layout
    plt.tight_layout()

    return fig, ax


def centroid_spectrum(mz_list, intensity_list, centroid_mode='max',
                      ms2_da=0.005, ms2_ppm=25.0):
    """
    Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da.
    centroid_mode: 'max' or 'sum'
    """
    peaks = list(zip(mz_list, intensity_list))
    peaks = np.asarray(peaks, dtype=np.float32, order="C")
    # Sort the peaks by m/z.
    peaks = peaks[np.argsort(peaks[:, 0])]
    peaks = _centroid_spectrum(peaks, centroid_mode, ms2_da=ms2_da, ms2_ppm=ms2_ppm)

    return peaks[:, 0], peaks[:, 1]


@njit
def _centroid_spectrum(peaks, centroid_mode='max',
                       ms2_da=0.005, ms2_ppm=25.0) -> np.ndarray:
    """Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da,
    only merging peaks with lower intensity than the target peak."""
    # Construct a new spectrum to avoid memory reallocation.
    peaks_new = np.zeros((peaks.shape[0], 2), dtype=np.float32)
    peaks_new_len = 0

    # Get the intensity argsort order.
    intensity_order = np.argsort(peaks[:, 1])

    # Iterate through the peaks from high to low intensity.
    for idx in intensity_order[::-1]:
        if peaks[idx, 1] > 0:
            mz_delta_allowed_ppm_in_da = peaks[idx, 0] * ms2_ppm * 1e-6
            mz_delta_allowed = max(ms2_da, mz_delta_allowed_ppm_in_da)

            # Find indices of peaks to merge.
            indices = np.where(np.logical_and(np.abs(peaks[:, 0] - peaks[idx, 0]) <= mz_delta_allowed,
                                              peaks[:, 1] < peaks[idx, 1]))[0]

            if centroid_mode == 'max':
                # Merge the peaks
                peaks_new[peaks_new_len, 0] = peaks[idx, 0]
                peaks_new[peaks_new_len, 1] = peaks[idx, 1]
                peaks_new_len += 1

                peaks[indices, 1] = 0

            elif centroid_mode == 'sum':
                # Merge the peaks
                intensity_sum = peaks[idx, 1]
                intensity_weighted_mz_sum = peaks[idx, 1] * peaks[idx, 0]
                for i in indices:
                    intensity_sum += peaks[i, 1]
                    intensity_weighted_mz_sum += peaks[i, 1] * peaks[i, 0]
                    peaks[i, 1] = 0  # Set the intensity of the merged peaks to 0

                peaks_new[peaks_new_len, 0] = intensity_weighted_mz_sum / intensity_sum
                peaks_new[peaks_new_len, 1] = intensity_sum
                peaks_new_len += 1

                # Set the intensity of the target peak to 0
                peaks[idx, 1] = 0

    # Return the new spectrum.
    peaks_new = peaks_new[:peaks_new_len]
    return peaks_new[np.argsort(peaks_new[:, 0])]


def _check_centroid(peaks, ms2_da=0.005, ms2_ppm=25.0) -> int:
    """Check if the spectrum is centroided. 0 for False and 1 for True."""
    if peaks.shape[0] <= 1:
        return 1

    # Calculate ms2_ppm_in_da
    ms2_ppm_in_da = peaks[1:, 0] * ms2_ppm * 1e-6

    # Use bitwise OR to choose the maximum
    threshold = np.maximum(ms2_da, ms2_ppm_in_da)

    # Check if the spectrum is centroided
    return 1 if np.all(np.diff(peaks[:, 0]) >= threshold) else 0


def plot_filtered_spectrum(filtered_mz, filtered_intensity, noise_mz, noise_ints,
                           title=None, x_range=None, y_range=None):
    import matplotlib.pyplot as plt

    # Create the plot
    fig, ax = plt.subplots(figsize=(15, 6))

    # Plot the filtered spectrum
    ax.vlines(filtered_mz, [0], filtered_intensity, linewidth=1, label='Filtered Spectrum')

    # Plot the noise
    ax.vlines(noise_mz, [0], noise_ints, color='red', linewidth=1, alpha=0.5, label='Noise')

    # Set labels and title
    ax.set_xlabel('m/z')
    ax.set_ylabel('Intensity')

    # Set axis limits if provided
    if x_range:
        ax.set_xlim(x_range[0], x_range[1])
    if y_range:
        ax.set_ylim(y_range[0], y_range[1])

    # Add grid
    ax.grid(True, linestyle='--', alpha=0.7)

    # Add legend
    ax.legend()

    # Add title
    if title:
        ax.set_title(title)

    # Adjust layout
    plt.tight_layout()

    return fig, ax


if __name__ == "__main__":
    # imzml_file = '/Users/shipei/Documents/projects/ms1_id/imaging/HeLa/14052018_coculture_HeLa_NIH3T3_H3_Neg.imzML'

    imzml_file = '/Users/shipei/Documents/projects/ms1_id/imaging/macrophage/20220127_250x271_5um_single cell macro eau_AcOH_dan_Neg_mode_200-2000mz_70K_Laser39_6.5kV_400C_Slens90_Line_leftright_centroid.imzML'

    mz_array, intensity_array = get_spectrum_from_imzml(imzml_file, index=10)
    peaks = np.column_stack((mz_array, intensity_array))

    filtered_mz, filtered_intensity, noise_mz, noise_ints = apply_moving_average_numba(mz_array, intensity_array,
                                                                                       mz_window=100.0,
                                                                                       percentage_lowest=0.05,
                                                                                       factor=1)

    centroid_mz, centroid_intensity = centroid_spectrum(filtered_mz, filtered_intensity,
                                                        centroid_mode='max',
                                                        ms2_da=0.005, ms2_ppm=25)

    x_range = None
    y_range = None

    # mask = (peaks[:, 0] >= x_range[0]) & (peaks[:, 0] <= x_range[1])
    # print(peaks[mask])

    fig, ax = plot_ms_spectrum(mz_array, intensity_array, title='original', x_range=x_range, y_range=y_range)
    fig, ax = plot_filtered_spectrum(filtered_mz, filtered_intensity, noise_mz, noise_ints,
                                     title='filtered', x_range=x_range, y_range=y_range)
    fig, ax = plot_ms_spectrum(centroid_mz, centroid_intensity, title='centroided', x_range=x_range, y_range=y_range)

    # Display the plot
    plt.show()

    # Optionally, save the plot
    # fig.savefig('ms_spectrum_pixel_0.png', dpi=300, bbox_inches='tight')



"""
Annotate pseudo MS/MS spectra in MGF format using indexed MS/MS libraries.
"""

import os
import pickle

import numpy as np
import pandas as pd
from ms_entropy import read_one_spectrum
from tqdm import tqdm

from ms1_id.lcms.centroid_data import centroid_spectrum_for_search


def annotate_pseudo_ms2_spec(ms2_file_path, library_ls,
                             mz_tol=0.05,
                             ion_mode=None,
                             score_cutoff=0.7,
                             min_matched_peak=3,
                             min_spec_usage=0.20,
                             save_dir=None):
    """
    Annotate pseudo MS/MS spectra in MGF format using indexed MS/MS libraries.

    :param ms2_file_path: path to the MGF file containing pseudo MS2 spectra
    :param library_ls: list of paths to indexed MS/MS libraries
    :param mz_tol: m/z tolerance in Da, for open matching
    :param ion_mode: ion mode, 'positive' or 'negative' or None
    :param score_cutoff: minimum score for matching
    :param min_matched_peak: minimum number of matched peaks
    :param min_spec_usage: minimum spectral usage
    :param save_dir: path to save the annotated results
    :return: DataFrame containing the annotation results
    """
    mz_tol = min(mz_tol, 0.05)  # indexed library mz_tol is 0.05

    # if library_ls is a string, convert to list
    if isinstance(library_ls, str):
        library_ls = [library_ls]

    all_matches = []

    # Load db
    for library in library_ls:
        with open(library, 'rb') as file:
            search_eng = pickle.load(file)
        db_name = os.path.basename(library)
        print(f'Loaded {db_name}')

        spec_idx = 0
        for spec in tqdm(read_one_spectrum(ms2_file_path)):

            spec_idx += 1

            # qry spec id
            spec_id = spec.get('title', None) or spec.get('spectrum_id', None) or spec.get('spectrumid', None) or \
                      spec.get('feature_id', None) or spec.get('featureid', None) or spec.get('scans', None)
            if spec_id is None:
                spec_id = spec_idx

            # qry precursor mz
            spec_mz = spec.get('precursor_mz', None) or spec.get('precursormz', None) or spec.get('pepmass', None)

            # skip if no peaks
            if not spec.get('peaks', None):
                continue

            peaks = np.asarray(spec['peaks'], dtype=np.float32, order="C")
            peaks = centroid_spectrum_for_search(peaks, width_da=0.05 * 2.015)

            if len(peaks) < min_matched_peak:
                continue

            matching_result = search_eng.search(
                precursor_mz=2000.00,  # unused, open search
                peaks=peaks,
                ms1_tolerance_in_da=mz_tol,
                ms2_tolerance_in_da=mz_tol,
                method="open",
                precursor_ions_removal_da=0.5,
                noise_threshold=0.0,
                min_ms2_difference_in_da=mz_tol * 2.02,
                reverse=True
            )

            score_arr, matched_peak_arr, spec_usage_arr = matching_result['open_search']

            # filter by matching cutoffs
            v = np.where((score_arr >= score_cutoff) &
                         (matched_peak_arr >= min_matched_peak) &
                         (spec_usage_arr >= min_spec_usage))[0]

            if len(v) == 0:
                continue

            for idx in v:
                matched = {k.lower(): v for k, v in search_eng[idx].items()}

                this_ion_mode = matched.get('ion_mode', '')
                if ion_mode is not None and ion_mode != this_ion_mode:
                    continue

                # precursor should be in the pseudo MS2 spectrum
                precursor_mz = matched.get('precursor_mz', 0)
                if not any(np.isclose(peaks[:, 0], precursor_mz, atol=mz_tol)):
                    continue

                all_matches.append({
                    'id': spec_id,
                    'mz': spec_mz,
                    'score': score_arr[idx],
                    'matched_peaks': matched_peak_arr[idx],
                    'spectral_usage': spec_usage_arr[idx],
                    'ref_name': matched.get('name', None),
                    'ref_precursor_mz': matched.get('precursor_mz', None),
                    'ref_precursor_type': matched.get('precursor_type', None),
                    'ref_formula': matched.get('formula', None),
                    'ref_inchikey': matched.get('inchikey', None),
                    'ref_instrument_type': matched.get('instrument_type', None),
                    'ref_collision_energy': matched.get('collision_energy', None),
                    'ref_comment': matched.get('comment', None),
                    'ref_db_name': db_name
                })

    df = pd.DataFrame(all_matches)

    # sort by id, then score
    df_top1 = df.sort_values(by=['id', 'score'], ascending=[True, False]).reset_index(drop=True)
    df_top1 = df_top1.groupby('id').head(1)

    out_basename = os.path.splitext(os.path.basename(ms2_file_path))[0] + '_annotations_all.tsv'
    out_top1_basename = os.path.splitext(os.path.basename(ms2_file_path))[0] + '_annotations_top1.tsv'
    if save_dir is not None:
        out_dir = save_dir
    else:
        # folder of the input file
        out_dir = os.path.dirname(ms2_file_path)

    df.to_csv(os.path.join(out_dir, out_basename), sep='\t', index=False)
    df_top1.to_csv(os.path.join(out_dir, out_top1_basename), sep='\t', index=False)

    return


if __name__ == "__main__":
    annotate_pseudo_ms2_spec(
        '/demo/lc_ms/single_files/NIST_pool_1_20eV_pseudo_ms2.mgf',
        ['/Users/shipei/Documents/projects/ms1_id/data/gnps.pkl',
         '/Users/shipei/Documents/projects/ms1_id/data/gnps_k10.pkl'])



"""
This file contains code modified from the FlashEntropySearch project
(https://github.com/YuanyueLi/FlashEntropySearch)
Copyright Yuanyue Li in 2023

Modified by Shipei Xing in 2024

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from numba import njit


def centroid_spectrum(mz_list, intensity_list, centroid_mode='max',
                      width_da=0.005, width_ppm=25.0):
    """
    Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da.
    centroid_mode: 'max' or 'sum'
    """
    peaks = list(zip(mz_list, intensity_list))
    peaks = np.asarray(peaks, dtype=np.float32, order="C")

    if len(peaks) == 0:
        return mz_list, intensity_list

    # Sort the peaks by m/z.
    peaks = peaks[np.argsort(peaks[:, 0])]

    # centroid the spectrum
    peaks = _centroid_spectrum(peaks, centroid_mode, width_da=width_da, width_ppm=width_ppm)
    return peaks[:, 0], peaks[:, 1]


@njit
def _centroid_spectrum(peaks, centroid_mode='max', peak_height_requirement=True,
                       width_da=0.005, width_ppm=25.0) -> np.ndarray:
    """Centroid a spectrum by merging peaks within the +/- ms2_ppm or +/- ms2_da,
    only merging peaks with lower intensity than the target peak."""
    # Construct a new spectrum to avoid memory reallocation.
    peaks_new = np.zeros((peaks.shape[0], 2), dtype=np.float32)
    peaks_new_len = 0

    # Get the intensity argsort order.
    intensity_order = np.argsort(peaks[:, 1])

    # Iterate through the peaks from high to low intensity.
    for idx in intensity_order[::-1]:
        if peaks[idx, 1] > 0:
            mz_delta_allowed_ppm_in_da = peaks[idx, 0] * width_ppm * 1e-6
            mz_delta_allowed = max(width_da, mz_delta_allowed_ppm_in_da)

            # Find indices of peaks to merge.
            if peak_height_requirement:
                indices = np.where(np.logical_and(np.abs(peaks[:, 0] - peaks[idx, 0]) <= mz_delta_allowed,
                                                  peaks[:, 1] < peaks[idx, 1]))[0]
            else:
                indices = np.where(np.abs(peaks[:, 0] - peaks[idx, 0]) <= mz_delta_allowed)[0]

            if centroid_mode == 'max':
                # Merge the peaks
                peaks_new[peaks_new_len, 0] = peaks[idx, 0]
                peaks_new[peaks_new_len, 1] = peaks[idx, 1]
                peaks_new_len += 1

                peaks[indices, 1] = 0

            elif centroid_mode == 'sum':
                # Merge the peaks
                intensity_sum = peaks[idx, 1]
                intensity_weighted_mz_sum = peaks[idx, 1] * peaks[idx, 0]
                for i in indices:
                    intensity_sum += peaks[i, 1]
                    intensity_weighted_mz_sum += peaks[i, 1] * peaks[i, 0]
                    peaks[i, 1] = 0  # Set the intensity of the merged peaks to 0

                peaks_new[peaks_new_len, 0] = intensity_weighted_mz_sum / intensity_sum
                peaks_new[peaks_new_len, 1] = intensity_sum
                peaks_new_len += 1

                # Set the intensity of the target peak to 0
                peaks[idx, 1] = 0

    # Return the new spectrum.
    peaks_new = peaks_new[:peaks_new_len]
    return peaks_new[np.argsort(peaks_new[:, 0])]


def _check_centroid(peaks, width_da=0.005, width_ppm=25.0) -> int:
    """Check if the spectrum is centroided. 0 for False and 1 for True."""
    if peaks.shape[0] <= 1:
        return 1

    # Calculate width_ppm_in_da
    width_ppm_in_da = peaks[1:, 0] * width_ppm * 1e-6

    # Use bitwise OR to choose the maximum
    threshold = np.maximum(width_da, width_ppm_in_da)

    # Check if the spectrum is centroided
    return 1 if np.all(np.diff(peaks[:, 0]) >= threshold) else 0


def check_centroid_for_search(peaks, width_da=0.105):
    """Check if the spectrum is centroided. True for centroided and False for not centroided."""

    # Check if the spectrum is centroided
    return True if np.all(np.diff(peaks[:, 0]) >= width_da) else False


def centroid_spectrum_for_search(peaks, width_da=0.105):
    """Centroid a spectrum for search."""
    if len(peaks) == 0:
        return peaks

    if check_centroid_for_search(peaks, width_da=width_da):
        return peaks

    # Sort the peaks by m/z.
    peaks = peaks[np.argsort(peaks[:, 0])]

    # centroid the spectrum
    peaks = _centroid_spectrum(peaks, centroid_mode='max', width_da=width_da)
    return peaks



import pandas as pd
from tqdm import tqdm

import json


def load_workbench_df():

    with open('/data/PR000639/metadata/workbench_db/all', 'r') as file:
        data = json.load(file)

    # convert the data to a list of dictionaries
    rows = [value for key, value in data.items()]

    df = pd.DataFrame(rows)

    print(df.head())

    df.to_csv('workbench_db.tsv', sep='\t', index=False)


def add_inchikey_to_workbench_results():

    df = pd.read_csv('/data/PR000639/metadata/workbench_identifications.tsv',
                     sep='\t', low_memory=False)

    meta = pd.read_csv('workbench_db.tsv', sep='\t', low_memory=False)

    # dict from name to inchikey
    name_to_inchikey = dict(zip(meta['name'], meta['inchi_key']))

    # add inchikey to df
    df['inchikey'] = df['refmet_name'].map(name_to_inchikey)

    df.to_csv('workbench_identifications_with_inchikey.tsv', sep='\t', index=False)


def filter_stats_df():
    df = pd.read_csv('all_modes_statistical_analysis.tsv', sep='\t', low_memory=False)
    print(f"Original shape: {df.shape}")

    # remove all cols starting with 'PREF' (pool reference mixture)
    df = df.loc[:, ~df.columns.str.startswith('PREF')]

    # filter out rows with score and matched peaks
    df = df[(df['MS1_similarity'] >= 0.80) & (df['MS1_matched_peak'] >= 6)].reset_index(drop=True)

    # sort by MS1 inchikey, then by MS1_similarity
    df = df.sort_values(['MS1_inchikey', 'MS1_similarity'], ascending=[True, False]).reset_index(drop=True)

    # remove duplicated MS1 inchikey
    df = df.drop_duplicates('MS1_inchikey', keep='first').reset_index(drop=True)

    ########################################################
    # remove rows which show up in the workbench results
    workbench_df = pd.read_csv('workbench_identifications_with_inchikey.tsv', sep='\t', low_memory=False)
    workbench_inchikeys = workbench_df['inchikey'].tolist()
    df = df[~df['MS1_inchikey'].isin(workbench_inchikeys)].reset_index(drop=True)

    ########################################################
    # filter rows by corrected p values
    df = df[(df['nonIBD_vs_CD_p_value_corrected'] <= 0.05) | (df['nonIBD_vs_UC_p_value_corrected'] <= 0.05)].reset_index(drop=True)

    ########################################################
    # filter rows by fill_percentage
    df = df[df['fill_percentage'] >= 0.25].reset_index(drop=True)

    print(f"Filtered shape: {df.shape}")
    df.to_csv('filtered_statistical_analysis.tsv', sep='\t', index=False)


if __name__ == '__main__':
    # load_workbench_df()

    # add_inchikey_to_workbench_results()

    filter_stats_df()

    # manual selection: filtered_statistical_analysis_manual.tsv






import os
import pandas as pd
import numpy as np
from ms1_id.lcms.utils import AlignedMS1Annotation, AnnotatedPseudoMS2
from ms1_id.lcms.centroid_data import consensus_spectrum


# def write_ms1_id_results(ms1_spec_ls, out_dir=None):
#     """
#     Output the annotated ms1 spectra
#     :param ms1_spec_ls: a list of PseudoMS2-like object
#     :param out_dir: output folder
#     :return: None
#     """
#
#     # only write out spectra with annotations
#     ms1_spec_ls = [spec for spec in ms1_spec_ls if spec.annotated]
#
#     out_list = []
#     for spec in ms1_spec_ls:
#         # pseudo_ms2_str: mz1 int1; mz2 int2; ...
#         pseudo_ms2_str = ''.join([f"{mz:.4f} {intensity:.0f};" for mz, intensity in zip(spec.mzs, spec.intensities)])
#
#         for annotation in spec.annotation_ls:
#             out_list.append({
#                 'file_name': spec.file_name,
#                 'rt': round(spec.rt, 2) if spec.rt else None,
#                 'db_name': annotation.db_name,
#                 'name': annotation.name,
#                 'precursor_mz': round(annotation.precursor_mz, 4),
#                 'matched_score': round(annotation.score, 4),
#                 'matched_peak': annotation.matched_peak,
#                 'spectral_usage': round(annotation.spectral_usage, 4) if annotation.spectral_usage else None,
#                 # 'search_eng_matched_id': annotation.search_eng_matched_id,
#                 'precursor_type': annotation.precursor_type,
#                 'formula': annotation.formula,
#                 'inchikey': annotation.inchikey,
#                 'instrument_type': annotation.instrument_type,
#                 'collision_energy': annotation.collision_energy,
#                 'db_id': annotation.db_id,
#                 'pseudo_ms2': pseudo_ms2_str,
#             })
#
#     out_df = pd.DataFrame(out_list)
#
#     out_path = os.path.join(out_dir, 'ms1_id_results.tsv')
#     out_df.to_csv(out_path, index=False, sep='\t')
#
#     return out_df


def write_single_file(msdata, pseudo_ms2_spectra=None, save_path=None):
    """
    Function to generate a report for rois in csv format.
    """

    result = []

    for roi in msdata.rois:
        iso_dist = ""
        for i in range(len(roi.isotope_mz_seq)):
            iso_dist += (str(np.round(roi.isotope_mz_seq[i], decimals=4)) + ";" +
                         str(np.round(roi.isotope_int_seq[i], decimals=0)) + "|")
        iso_dist = iso_dist[:-1]

        ms2 = ""
        if roi.best_ms2 is not None:
            for i in range(len(roi.best_ms2.peaks)):
                ms2 += (str(np.round(roi.best_ms2.peaks[i, 0], decimals=4)) + ";" +
                        str(np.round(roi.best_ms2.peaks[i, 1], decimals=0)) + "|")
            ms2 = ms2[:-1]

        temp = [roi.id, roi.mz.__round__(4), roi.rt.__round__(3), roi.length, roi.rt_seq[0],
                roi.rt_seq[-1], roi.peak_area, roi.peak_height, roi.gaussian_similarity.__round__(2),
                roi.noise_level.__round__(2), roi.asymmetry_factor.__round__(2), roi.charge_state, roi.is_isotope,
                str(roi.isotope_id_seq)[1:-1], iso_dist,
                # roi.is_in_source_fragment, roi.isf_parent_roi_id, str(roi.isf_child_roi_id)[1:-1],
                roi.adduct_type,
                # roi.adduct_parent_roi_id, str(roi.adduct_child_roi_id)[1:-1],
                ]

        temp.extend([ms2, roi.annotation, roi.formula, roi.similarity, roi.matched_peak_number,
                     roi.inchikey, getattr(roi, 'precursor_type', None), getattr(roi, 'collision_energy', None)])

        result.append(temp)

    # convert result to a pandas dataframe
    columns = ["ID", "m/z", "RT", "length", "RT_start", "RT_end", "peak_area", "peak_height",
               "Gaussian_similarity", "noise_level", "asymmetry_factor", "charge", "is_isotope",
               "isotope_IDs", "isotopes",
               # "is_in_source_fragment", "ISF_parent_ID", "ISF_child_ID",
               "adduct",
               # "adduct_base_ID", "adduct_other_ID"
               ]

    columns.extend(["MS2", "MS2_annotation", "MS2_formula", "MS2_similarity", "MS2_matched_peak",
                    "MS2_inchikey", "MS2_precursor_type", "MS2_collision_energy"])

    df = pd.DataFrame(result, columns=columns)

    # add ms1 id results
    df['MS1_annotation'] = None
    df['MS1_formula'] = None
    df['MS1_similarity'] = None
    df['MS1_matched_peak'] = None
    df['MS1_spectral_usage'] = None
    df['MS1_precursor_type'] = None
    df['MS1_inchikey'] = None
    df['MS1_collision_energy'] = None
    df['MS1_db_name'] = None
    df['MS1_db_id'] = None
    df['pseudo_ms2'] = None

    if pseudo_ms2_spectra is not None:
        spec_ls = [spec for spec in pseudo_ms2_spectra if spec.annotated]

        if len(spec_ls) > 0:
            for spec in spec_ls:

                pseudo_ms2_str = ''.join(
                    [f"{mz:.4f} {intensity:.0f};" for mz, intensity in zip(spec.mzs, spec.intensities)])

                for annotation in spec.annotation_ls:
                    this_precmz = annotation.precursor_mz
                    # Create a boolean mask for the conditions
                    mask = (((df['m/z'] - this_precmz).abs() <= msdata.params.mz_tol_ms1) &
                            ((df['RT'] - spec.rt).abs() <= 0.1))

                    if not mask.any():
                        continue

                    idx = df.loc[mask, 'RT'].sub(spec.rt).abs().idxmin()

                    # if the annotation is better than the previous one, update the row
                    if df.loc[idx, 'MS1_similarity'] is None or annotation.score >= df.loc[idx, 'MS1_similarity']:
                        df.loc[idx, 'MS1_annotation'] = annotation.name
                        df.loc[idx, 'MS1_formula'] = annotation.formula
                        df.loc[idx, 'MS1_similarity'] = round(float(annotation.score), 4)
                        df.loc[idx, 'MS1_matched_peak'] = annotation.matched_peak
                        df.loc[idx, 'MS1_spectral_usage'] = round(float(annotation.spectral_usage), 4)
                        df.loc[idx, 'MS1_precursor_type'] = annotation.precursor_type
                        df.loc[idx, 'MS1_inchikey'] = annotation.inchikey
                        df.loc[idx, 'MS1_collision_energy'] = annotation.collision_energy
                        df.loc[idx, 'MS1_db_name'] = annotation.db_name
                        df.loc[idx, 'MS1_db_id'] = annotation.db_id
                        df.loc[idx, 'pseudo_ms2'] = pseudo_ms2_str

        # # for unannotated features, fill in the pseudo MS2 spectra
        # df = fill_pseudo_ms2_spectra_for_unannotated_features(df, pseudo_ms2_spectra, msdata.params)

    # save the dataframe to csv file
    df.to_csv(save_path, index=False, sep="\t")


def write_feature_table(df, pseudo_ms2_spectra, config, output_path):
    """
    A function to output the aligned feature table.
    :param df: pd.DataFrame, feature table
    :param pseudo_ms2_spectra: list of PseudoMS2-like object
    :param config: Parameters
    :param output_path: str, output file path
        columns=["ID", "m/z", "RT", "adduct", "is_isotope", "is_in_source_fragment", "Gaussian_similarity", "noise_level",
             "asymmetry_factor", "charge", "isotopes", "MS2", "matched_MS2", "search_mode",
             "annotation", "formula", "similarity", "matched_peak_number", "SMILES", "InChIKey",
                "fill_percentage", "alignment_reference"] + sample_names
    """

    # keep four digits for the m/z column and three digits for the RT column
    df["m/z"] = df["m/z"].apply(lambda x: round(x, 4))
    df["RT"] = df["RT"].apply(lambda x: round(x, 3))
    df['fill_percentage'] = df['fill_percentage'].apply(lambda x: round(x, 2))
    df['similarity'] = df['similarity'].astype(float)
    df['similarity'] = df['similarity'].apply(lambda x: round(x, 4))

    # refine ms1 id results with feature table. for each feature, choose the most confident annotation
    aligned_ms1_annotation_ls = refine_pseudo_ms2_spectra_list(pseudo_ms2_spectra, df, config)

    # Add MS1 id results columns
    ms1_columns = ['MS1_annotation', 'MS1_formula', 'MS1_similarity', 'MS1_matched_peak',
                   'MS1_spectral_usage', 'MS1_precursor_type', 'MS1_inchikey',
                   'MS1_collision_energy', 'MS1_db_name', 'MS1_db_id', 'pseudo_ms2']
    for col in ms1_columns:
        df[col] = None

    # add ms1 id results to the feature table
    df = add_ms1_id_results(df, aligned_ms1_annotation_ls)

    # for unannotated features, fill in the pseudo MS2 spectra
    df = fill_pseudo_ms2_spectra_for_unannotated_features(df, pseudo_ms2_spectra, config)

    df.to_csv(output_path, index=False, sep="\t")


def add_ms1_id_results(df, aligned_ms1_annotation_ls):
    """
    Add MS1 ID results to the feature table
    """
    for aligned_ms1_annotation in aligned_ms1_annotation_ls:
        idx = aligned_ms1_annotation.df_idx

        annotation = aligned_ms1_annotation.selected_annotated_pseudo_ms2.annotation
        df.loc[idx, 'MS1_annotation'] = annotation.name
        df.loc[idx, 'MS1_formula'] = annotation.formula
        df.loc[idx, 'MS1_similarity'] = round(float(annotation.score), 4)
        df.loc[idx, 'MS1_matched_peak'] = annotation.matched_peak
        df.loc[idx, 'MS1_spectral_usage'] = round(float(annotation.spectral_usage), 4)
        df.loc[idx, 'MS1_precursor_type'] = annotation.precursor_type
        df.loc[idx, 'MS1_inchikey'] = annotation.inchikey
        df.loc[idx, 'MS1_collision_energy'] = annotation.collision_energy
        df.loc[idx, 'MS1_db_name'] = annotation.db_name
        df.loc[idx, 'MS1_db_id'] = annotation.db_id

        pseudo_ms2_str = ''.join(
            [f"{mz:.4f} {intensity:.0f};" for mz, intensity in
             zip(aligned_ms1_annotation.selected_annotated_pseudo_ms2.pseudo_ms2_mzs,
                 aligned_ms1_annotation.selected_annotated_pseudo_ms2.pseudo_ms2_intensities)])
        df.loc[idx, 'pseudo_ms2'] = pseudo_ms2_str

    return df


def fill_pseudo_ms2_spectra_for_unannotated_features(df, pseudo_ms2_spectra, config):
    """
    Fill in the pseudo MS2 spectra for unannotated features
    """
    for spec in pseudo_ms2_spectra:
        # Create a boolean mask for the conditions
        mask = (((df['m/z'] - spec.t_mz).abs() <= config.align_mz_tol) &
                ((df['RT'] - spec.rt).abs() <= config.align_rt_tol))

        if not mask.any():
            continue

        # Find the row with the smallest RT difference
        idx = df.loc[mask, 'RT'].sub(spec.rt).abs().idxmin()

        # if already annotated, skip
        if df.loc[idx, 'MS1_similarity'] is not None:
            continue

        pseudo_ms2_peaks = np.column_stack((np.array(spec.mzs), np.array(spec.intensities)))

        if df.loc[idx, 'pseudo_ms2'] is None:
            df.loc[idx, 'pseudo_ms2'] = [pseudo_ms2_peaks]
        else:
            df.loc[idx, 'pseudo_ms2'].append(pseudo_ms2_peaks)

    # consensus pseudo MS2 spectra for unannotated features
    for idx, row in df.iterrows():
        if row['MS1_similarity'] is None and row['pseudo_ms2'] is not None:
            pseudo_ms2_peaks_ls = row['pseudo_ms2']

            merged_peaks = np.concatenate(pseudo_ms2_peaks_ls, axis=0)

            # find the consensus pseudo MS2 spectrum
            consensus_peaks = consensus_spectrum(merged_peaks, config.mz_tol_ms2)

            pseudo_ms2_str = ''.join([f"{mz:.4f} {intensity:.0f};" for mz, intensity in consensus_peaks])

            # add the consensus pseudo MS2 spectrum to the feature table
            df.loc[idx, 'pseudo_ms2'] = pseudo_ms2_str

    return df


def refine_pseudo_ms2_spectra_list(pseudo_ms2_spectra, df, config):
    """
    for each feature, choose the most confident annotation
    :return: AlignedMS1Annotation with selected annotations
    """

    # only reserve annotated pseudo MS2 spectra
    spec_ls = [spec for spec in pseudo_ms2_spectra if spec.annotated]

    all_df_idx_ls = []  # list of indices in the feature table that have been matched
    aligned_ms1_annotation_ls = []  # list of AlignedMS1Annotation objects

    for spec in spec_ls:
        this_rt = spec.rt
        for annotation in spec.annotation_ls:
            this_precmz = annotation.precursor_mz
            # Create a boolean mask for the conditions
            mask = ((df['m/z'] - this_precmz).abs() <= config.align_mz_tol) & (
                        (df['RT'] - this_rt).abs() <= config.align_rt_tol)

            if not mask.any():
                continue

            # Find the row with the smallest RT difference
            idx = df.loc[mask, 'RT'].sub(this_rt).abs().idxmin()

            annotated_pseudo_ms2 = AnnotatedPseudoMS2(annotation, spec.mzs, spec.intensities)

            if idx in all_df_idx_ls:
                aligned_ms1_annotation_idx = all_df_idx_ls.index(idx)
                # add annotation to the existing AlignedMS1Annotation object
                aligned_ms1_annotation_ls[aligned_ms1_annotation_idx].annotated_pseudo_ms2_list.append(
                    annotated_pseudo_ms2)
            else:
                # add the index to the list
                all_df_idx_ls.append(idx)
                # create an AlignedMS1Annotation object
                aligned_ms1_annotation = AlignedMS1Annotation(idx)
                aligned_ms1_annotation.annotated_pseudo_ms2_list.append(annotated_pseudo_ms2)

                # add the object to the total list
                aligned_ms1_annotation_ls.append(aligned_ms1_annotation)

    # find the one with the highest similarity score
    for aligned_ms1_annotation in aligned_ms1_annotation_ls:
        if len(aligned_ms1_annotation.annotated_pseudo_ms2_list) > 1:
            aligned_ms1_annotation.selected_annotated_pseudo_ms2 = (
                max(aligned_ms1_annotation.annotated_pseudo_ms2_list, key=lambda x: x.annotation.score))
        else:
            aligned_ms1_annotation.selected_annotated_pseudo_ms2 = aligned_ms1_annotation.annotated_pseudo_ms2_list[0]

    return aligned_ms1_annotation_ls


def write_pseudoms2_to_mgf(pseudoms2_ls, save_dir, file_name):
    """
    Write pseudo MS2 spectra to MGF file
    """

    mgf_path = os.path.join(save_dir, f"{file_name}_pseudo_ms2.mgf")

    with open(mgf_path, 'w') as f:
        idx = 1
        for spec in pseudoms2_ls:
            f.write(f"BEGIN IONS\n")
            # f.write(f"PEPMASS={round(spec.t_mz, 5)}\n")
            f.write("PEPMASS=0")
            f.write(f"SCANS={idx}\n")
            f.write(f"RTINSECONDS={spec.rt * 60}\n")

            mz_arr = np.array(spec.mzs)
            intensity_arr = np.array(spec.intensities)

            # sort by mz
            sort_idx = np.argsort(mz_arr)
            mz_arr = mz_arr[sort_idx]
            intensity_arr = intensity_arr[sort_idx]

            for i in range(len(mz_arr)):
                f.write(f"{mz_arr[i]:.5f} {intensity_arr[i]:.0f}\n")

            f.write(f"END IONS\n\n")
            idx += 1
    return



from ms1_id.lcms.ms1id_lcms_single import main_workflow_single
from ms1_id.lcms.ms1id_lcms_batch import main_workflow
import os
import concurrent.futures
from functools import partial


def ms1id_single_file(file_path, ms1id_library_path=None, ms2id_library_path=None,
                      ms1_id=True, ms2_id=False,
                      ms1_tol=0.01, ms2_tol=0.015,
                      mass_detect_int_tol=10000,
                      peak_cor_rt_tol=0.025,
                      min_ppc=0.8, roi_min_length=4,
                      library_search_mztol=0.05,
                      ms1id_score_cutoff=0.7, ms1id_min_matched_peak=3, ms1id_min_spec_usage=0.05,
                      ms1id_max_prec_rel_int_in_other_ms2=0.01,
                      ms2id_score_cutoff=0.7, ms2id_min_matched_peak=3,
                      out_dir=None):
    print('===========================================')
    print(f'Processing {file_path}')

    main_workflow_single(
        file_path=file_path,
        ms1id_library_path=ms1id_library_path, ms2id_library_path=ms2id_library_path,
        ms1_id=ms1_id, ms2_id=ms2_id,
        ms1_tol=ms1_tol, ms2_tol=ms2_tol,
        mass_detect_int_tol=mass_detect_int_tol,
        peak_cor_rt_tol=peak_cor_rt_tol,
        min_ppc=min_ppc, roi_min_length=roi_min_length,
        library_search_mztol=library_search_mztol,
        ms1id_score_cutoff=ms1id_score_cutoff, ms1id_min_matched_peak=ms1id_min_matched_peak,
        ms1id_min_spec_usage=ms1id_min_spec_usage,
        ms1id_max_prec_rel_int_in_other_ms2=ms1id_max_prec_rel_int_in_other_ms2,
        ms2id_score_cutoff=ms2id_score_cutoff, ms2id_min_matched_peak=ms2id_min_matched_peak,
        out_dir=out_dir)

    return


def ms1id_single_file_batch(data_dir, ms1id_library_path=None, ms2id_library_path=None,
                            parallel=True, num_processes=None,
                            ms1_id=True, ms2_id=False,
                            ms1_tol=0.01, ms2_tol=0.015,
                            mass_detect_int_tol=10000,
                            peak_cor_rt_tol=0.025,
                            min_ppc=0.8, roi_min_length=4,
                            library_search_mztol=0.05,
                            ms1id_score_cutoff=0.7, ms1id_min_matched_peak=3, ms1id_min_spec_usage=0.05,
                            ms1id_max_prec_rel_int_in_other_ms2=0.01,
                            ms2id_score_cutoff=0.7, ms2id_min_matched_peak=3,
                            out_dir=None
                            ):
    """
    Process all files in a directory in single file mode using parallel processing.
    """
    # Get all mzXML and mzML files in the directory
    files = [f for f in os.listdir(data_dir) if f.endswith('.mzXML') or f.endswith('.mzML')]
    files = [os.path.join(data_dir, f) for f in files]

    # if some file results exist in out_dir, skip them
    if out_dir is not None:
        os.makedirs(out_dir, exist_ok=True)
        files = [f for f in files if not os.path.exists(
            os.path.join(out_dir, os.path.splitext(os.path.basename(f))[0] + '_feature_table.tsv'))]

    if len(files) == 0:
        print(f"No files to process in {data_dir}")
        return

    # Create a partial function with the library_path argument
    process_file = partial(ms1id_single_file,
                           ms1id_library_path=ms1id_library_path, ms2id_library_path=ms2id_library_path,
                           ms1_id=ms1_id, ms2_id=ms2_id,
                           ms1_tol=ms1_tol, ms2_tol=ms2_tol,
                           mass_detect_int_tol=mass_detect_int_tol,
                           peak_cor_rt_tol=peak_cor_rt_tol,
                           min_ppc=min_ppc, roi_min_length=roi_min_length,
                           library_search_mztol=library_search_mztol,
                           ms1id_score_cutoff=ms1id_score_cutoff, ms1id_min_matched_peak=ms1id_min_matched_peak,
                           ms1id_min_spec_usage=ms1id_min_spec_usage,
                           ms1id_max_prec_rel_int_in_other_ms2=ms1id_max_prec_rel_int_in_other_ms2,
                           ms2id_score_cutoff=ms2id_score_cutoff, ms2id_min_matched_peak=ms2id_min_matched_peak,
                           out_dir=out_dir)

    if parallel:
        # If num_processes is not specified, use the number of CPU cores
        if num_processes is None:
            num_processes = min(os.cpu_count(), len(files))

        with concurrent.futures.ProcessPoolExecutor(max_workers=num_processes) as executor:
            # Submit all tasks
            futures = [executor.submit(process_file, file) for file in files]

            # Wait for all tasks to complete
            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()  # This will raise any exceptions that occurred during execution
                except Exception as e:
                    print(f"An error occurred: {e}")
    else:
        # Process each file sequentially
        for file in files:
            ms1id_single_file(file_path=file,
                              ms1id_library_path=ms1id_library_path, ms2id_library_path=ms2id_library_path,
                              ms1_id=ms1_id, ms2_id=ms2_id,
                              ms1_tol=ms1_tol, ms2_tol=ms2_tol,
                              mass_detect_int_tol=mass_detect_int_tol,
                              peak_cor_rt_tol=peak_cor_rt_tol,
                              min_ppc=min_ppc, roi_min_length=roi_min_length,
                              library_search_mztol=library_search_mztol,
                              ms1id_score_cutoff=ms1id_score_cutoff, ms1id_min_matched_peak=ms1id_min_matched_peak,
                              ms1id_min_spec_usage=ms1id_min_spec_usage,
                              ms1id_max_prec_rel_int_in_other_ms2=ms1id_max_prec_rel_int_in_other_ms2,
                              ms2id_score_cutoff=ms2id_score_cutoff, ms2id_min_matched_peak=ms2id_min_matched_peak,
                              out_dir=out_dir)

    return


def ms1id_batch_mode(project_path=None, ms1id_library_path=None, ms2id_library_path=None,
                     sample_dir='data', parallel=True,
                     ms1_id=True, ms2_id=False,
                     cpu_ratio=0.8,
                     run_rt_correction=True, run_normalization=True,
                     ms1_tol=0.01, ms2_tol=0.015, mass_detect_int_tol=None,
                     align_mz_tol=0.015, align_rt_tol=0.2, alignment_drop_by_fill_pct_ratio=0.1,
                     peak_cor_rt_tol=0.025,
                     min_ppc=0.8, roi_min_length=5, library_search_mztol=0.05,
                     ms1id_score_cutoff=0.7, ms1id_min_matched_peak=3, ms1id_min_spec_usage=0.05,
                     ms1id_max_prec_rel_int_in_other_ms2=0.01,
                     ms2id_score_cutoff=0.7, ms2id_min_matched_peak=3):
    main_workflow(project_path=project_path, ms1id_library_path=ms1id_library_path,
                  ms2id_library_path=ms2id_library_path,
                  sample_dir=sample_dir,
                  parallel=parallel, ms1_id=ms1_id, ms2_id=ms2_id,
                  cpu_ratio=cpu_ratio,
                  run_rt_correction=run_rt_correction, run_normalization=run_normalization,
                  ms1_tol=ms1_tol, ms2_tol=ms2_tol, mass_detect_int_tol=mass_detect_int_tol,
                  align_mz_tol=align_mz_tol, align_rt_tol=align_rt_tol,
                  alignment_drop_by_fill_pct_ratio=alignment_drop_by_fill_pct_ratio,
                  peak_cor_rt_tol=peak_cor_rt_tol, min_ppc=min_ppc, roi_min_length=roi_min_length,
                  library_search_mztol=library_search_mztol,
                  ms1id_score_cutoff=ms1id_score_cutoff, ms1id_min_matched_peak=ms1id_min_matched_peak,
                  ms1id_min_spec_usage=ms1id_min_spec_usage,
                  ms1id_max_prec_rel_int_in_other_ms2=ms1id_max_prec_rel_int_in_other_ms2,
                  ms2id_score_cutoff=ms2id_score_cutoff, ms2id_min_matched_peak=ms2id_min_matched_peak)

    return
